from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf import wrappers_pb2 as _wrappers_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from qm.grpc.qm.pb import inc_qua_config_pb2 as _inc_qua_config_pb2
from google.protobuf import any_pb2 as _any_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class QuaProgram(_message.Message):
    __slots__ = ("config", "dynConfig", "script", "compilerOptions", "resultAnalysis", "config_update")
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        INT: _ClassVar[QuaProgram.Type]
        BOOL: _ClassVar[QuaProgram.Type]
        REAL: _ClassVar[QuaProgram.Type]
    INT: QuaProgram.Type
    BOOL: QuaProgram.Type
    REAL: QuaProgram.Type
    class Direction(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        INCOMING: _ClassVar[QuaProgram.Direction]
        OUTGOING: _ClassVar[QuaProgram.Direction]
    INCOMING: QuaProgram.Direction
    OUTGOING: QuaProgram.Direction
    class GlobalVarOperation(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        read: _ClassVar[QuaProgram.GlobalVarOperation]
        read_shift: _ClassVar[QuaProgram.GlobalVarOperation]
        xor: _ClassVar[QuaProgram.GlobalVarOperation]
    read: QuaProgram.GlobalVarOperation
    read_shift: QuaProgram.GlobalVarOperation
    xor: QuaProgram.GlobalVarOperation
    class CompilerOptions(_message.Message):
        __slots__ = ("useExperimentalCalculationCompiler", "optimizeMergeCodeExecution", "optimizeWriteReadCommands", "skipOptimizations", "strict", "flags")
        USEEXPERIMENTALCALCULATIONCOMPILER_FIELD_NUMBER: _ClassVar[int]
        OPTIMIZEMERGECODEEXECUTION_FIELD_NUMBER: _ClassVar[int]
        OPTIMIZEWRITEREADCOMMANDS_FIELD_NUMBER: _ClassVar[int]
        SKIPOPTIMIZATIONS_FIELD_NUMBER: _ClassVar[int]
        STRICT_FIELD_NUMBER: _ClassVar[int]
        FLAGS_FIELD_NUMBER: _ClassVar[int]
        useExperimentalCalculationCompiler: bool
        optimizeMergeCodeExecution: bool
        optimizeWriteReadCommands: bool
        skipOptimizations: _containers.RepeatedScalarFieldContainer[str]
        strict: bool
        flags: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, useExperimentalCalculationCompiler: bool = ..., optimizeMergeCodeExecution: bool = ..., optimizeWriteReadCommands: bool = ..., skipOptimizations: _Optional[_Iterable[str]] = ..., strict: bool = ..., flags: _Optional[_Iterable[str]] = ...) -> None: ...
    class Script(_message.Message):
        __slots__ = ("variables", "externalStreams", "body")
        VARIABLES_FIELD_NUMBER: _ClassVar[int]
        EXTERNALSTREAMS_FIELD_NUMBER: _ClassVar[int]
        BODY_FIELD_NUMBER: _ClassVar[int]
        variables: _containers.RepeatedCompositeFieldContainer[QuaProgram.VarDeclaration]
        externalStreams: _containers.RepeatedCompositeFieldContainer[QuaProgram.ExternalStreamDeclaration]
        body: QuaProgram.StatementsCollection
        def __init__(self, variables: _Optional[_Iterable[_Union[QuaProgram.VarDeclaration, _Mapping]]] = ..., externalStreams: _Optional[_Iterable[_Union[QuaProgram.ExternalStreamDeclaration, _Mapping]]] = ..., body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ...) -> None: ...
    class VarDeclaration(_message.Message):
        __slots__ = ("name", "type", "size", "value", "dim", "isInputStream", "structMember")
        class StructMember(_message.Message):
            __slots__ = ("name", "position")
            NAME_FIELD_NUMBER: _ClassVar[int]
            POSITION_FIELD_NUMBER: _ClassVar[int]
            name: str
            position: int
            def __init__(self, name: _Optional[str] = ..., position: _Optional[int] = ...) -> None: ...
        NAME_FIELD_NUMBER: _ClassVar[int]
        TYPE_FIELD_NUMBER: _ClassVar[int]
        SIZE_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        DIM_FIELD_NUMBER: _ClassVar[int]
        ISINPUTSTREAM_FIELD_NUMBER: _ClassVar[int]
        STRUCTMEMBER_FIELD_NUMBER: _ClassVar[int]
        name: str
        type: QuaProgram.Type
        size: int
        value: _containers.RepeatedCompositeFieldContainer[QuaProgram.LiteralExpression]
        dim: int
        isInputStream: bool
        structMember: QuaProgram.VarDeclaration.StructMember
        def __init__(self, name: _Optional[str] = ..., type: _Optional[_Union[QuaProgram.Type, str]] = ..., size: _Optional[int] = ..., value: _Optional[_Iterable[_Union[QuaProgram.LiteralExpression, _Mapping]]] = ..., dim: _Optional[int] = ..., isInputStream: bool = ..., structMember: _Optional[_Union[QuaProgram.VarDeclaration.StructMember, _Mapping]] = ...) -> None: ...
    class ExternalStreamDeclaration(_message.Message):
        __slots__ = ("stream_id", "expectedTypes", "direction")
        STREAM_ID_FIELD_NUMBER: _ClassVar[int]
        EXPECTEDTYPES_FIELD_NUMBER: _ClassVar[int]
        DIRECTION_FIELD_NUMBER: _ClassVar[int]
        stream_id: int
        expectedTypes: _containers.RepeatedCompositeFieldContainer[QuaProgram.VarDeclaration]
        direction: QuaProgram.Direction
        def __init__(self, stream_id: _Optional[int] = ..., expectedTypes: _Optional[_Iterable[_Union[QuaProgram.VarDeclaration, _Mapping]]] = ..., direction: _Optional[_Union[QuaProgram.Direction, str]] = ...) -> None: ...
    class QuantumElementReference(_message.Message):
        __slots__ = ("loc", "name")
        LOC_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        loc: str
        name: str
        def __init__(self, loc: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...
    class PulseReference(_message.Message):
        __slots__ = ("loc", "name")
        LOC_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        loc: str
        name: str
        def __init__(self, loc: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...
    class IntegrationWeightReference(_message.Message):
        __slots__ = ("loc", "name")
        LOC_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        loc: str
        name: str
        def __init__(self, loc: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...
    class AnyStatement(_message.Message):
        __slots__ = ("play", "measure", "wait", "sync", "assign", "rus", "align", "updateFrequency", "zRotation", "pause", "save", "forEach", "waitForTrigger", "qrun", "updateCorrection", "resetPhase", "rampToZero", "resetFrame", "setDcOffset", "advanceInputStream", "strictTiming", "fastFrameRotation", "updateOscillatorFrequency", "resetGlobalPhase", "loadWaveform", "sendToExternalStream", "receiveFromExternalStream", "globalVariableAssignment", "arbitrary", "arbitraryContext")
        PLAY_FIELD_NUMBER: _ClassVar[int]
        MEASURE_FIELD_NUMBER: _ClassVar[int]
        WAIT_FIELD_NUMBER: _ClassVar[int]
        SYNC_FIELD_NUMBER: _ClassVar[int]
        IF_FIELD_NUMBER: _ClassVar[int]
        ASSIGN_FIELD_NUMBER: _ClassVar[int]
        FOR_FIELD_NUMBER: _ClassVar[int]
        RUS_FIELD_NUMBER: _ClassVar[int]
        ALIGN_FIELD_NUMBER: _ClassVar[int]
        UPDATEFREQUENCY_FIELD_NUMBER: _ClassVar[int]
        ZROTATION_FIELD_NUMBER: _ClassVar[int]
        PAUSE_FIELD_NUMBER: _ClassVar[int]
        SAVE_FIELD_NUMBER: _ClassVar[int]
        FOREACH_FIELD_NUMBER: _ClassVar[int]
        WAITFORTRIGGER_FIELD_NUMBER: _ClassVar[int]
        QRUN_FIELD_NUMBER: _ClassVar[int]
        UPDATECORRECTION_FIELD_NUMBER: _ClassVar[int]
        RESETPHASE_FIELD_NUMBER: _ClassVar[int]
        RAMPTOZERO_FIELD_NUMBER: _ClassVar[int]
        RESETFRAME_FIELD_NUMBER: _ClassVar[int]
        SETDCOFFSET_FIELD_NUMBER: _ClassVar[int]
        ADVANCEINPUTSTREAM_FIELD_NUMBER: _ClassVar[int]
        STRICTTIMING_FIELD_NUMBER: _ClassVar[int]
        FASTFRAMEROTATION_FIELD_NUMBER: _ClassVar[int]
        UPDATEOSCILLATORFREQUENCY_FIELD_NUMBER: _ClassVar[int]
        RESETGLOBALPHASE_FIELD_NUMBER: _ClassVar[int]
        LOADWAVEFORM_FIELD_NUMBER: _ClassVar[int]
        SENDTOEXTERNALSTREAM_FIELD_NUMBER: _ClassVar[int]
        RECEIVEFROMEXTERNALSTREAM_FIELD_NUMBER: _ClassVar[int]
        GLOBALVARIABLEASSIGNMENT_FIELD_NUMBER: _ClassVar[int]
        ARBITRARY_FIELD_NUMBER: _ClassVar[int]
        ARBITRARYCONTEXT_FIELD_NUMBER: _ClassVar[int]
        play: QuaProgram.PlayStatement
        measure: QuaProgram.MeasureStatement
        wait: QuaProgram.WaitStatement
        sync: QuaProgram.SyncStatement
        assign: QuaProgram.AssignmentStatement
        rus: QuaProgram.RusStatement
        align: QuaProgram.AlignStatement
        updateFrequency: QuaProgram.UpdateFrequencyStatement
        zRotation: QuaProgram.ZRotationStatement
        pause: QuaProgram.PauseStatement
        save: QuaProgram.SaveStatement
        forEach: QuaProgram.ForEachStatement
        waitForTrigger: QuaProgram.WaitForTriggerStatement
        qrun: QuaProgram.QRunStatement
        updateCorrection: QuaProgram.UpdateCorrectionStatement
        resetPhase: QuaProgram.ResetPhaseStatement
        rampToZero: QuaProgram.RampToZeroStatement
        resetFrame: QuaProgram.ResetFrameStatement
        setDcOffset: QuaProgram.SetDcOffsetStatement
        advanceInputStream: QuaProgram.AdvanceInputStreamStatement
        strictTiming: QuaProgram.StrictTimingStatement
        fastFrameRotation: QuaProgram.FastFrameRotationStatement
        updateOscillatorFrequency: QuaProgram.UpdateOscillatorFrequencyStatement
        resetGlobalPhase: QuaProgram.ResetGlobalPhaseStatement
        loadWaveform: QuaProgram.LoadWaveformStatement
        sendToExternalStream: QuaProgram.SendToExternalStreamStatement
        receiveFromExternalStream: QuaProgram.ReceiveFromExternalStreamStatement
        globalVariableAssignment: QuaProgram.GlobalVariableAssignmentStatement
        arbitrary: QuaProgram.ArbitraryStatement
        arbitraryContext: QuaProgram.ArbitraryContextStatement
        def __init__(self, play: _Optional[_Union[QuaProgram.PlayStatement, _Mapping]] = ..., measure: _Optional[_Union[QuaProgram.MeasureStatement, _Mapping]] = ..., wait: _Optional[_Union[QuaProgram.WaitStatement, _Mapping]] = ..., sync: _Optional[_Union[QuaProgram.SyncStatement, _Mapping]] = ..., assign: _Optional[_Union[QuaProgram.AssignmentStatement, _Mapping]] = ..., rus: _Optional[_Union[QuaProgram.RusStatement, _Mapping]] = ..., align: _Optional[_Union[QuaProgram.AlignStatement, _Mapping]] = ..., updateFrequency: _Optional[_Union[QuaProgram.UpdateFrequencyStatement, _Mapping]] = ..., zRotation: _Optional[_Union[QuaProgram.ZRotationStatement, _Mapping]] = ..., pause: _Optional[_Union[QuaProgram.PauseStatement, _Mapping]] = ..., save: _Optional[_Union[QuaProgram.SaveStatement, _Mapping]] = ..., forEach: _Optional[_Union[QuaProgram.ForEachStatement, _Mapping]] = ..., waitForTrigger: _Optional[_Union[QuaProgram.WaitForTriggerStatement, _Mapping]] = ..., qrun: _Optional[_Union[QuaProgram.QRunStatement, _Mapping]] = ..., updateCorrection: _Optional[_Union[QuaProgram.UpdateCorrectionStatement, _Mapping]] = ..., resetPhase: _Optional[_Union[QuaProgram.ResetPhaseStatement, _Mapping]] = ..., rampToZero: _Optional[_Union[QuaProgram.RampToZeroStatement, _Mapping]] = ..., resetFrame: _Optional[_Union[QuaProgram.ResetFrameStatement, _Mapping]] = ..., setDcOffset: _Optional[_Union[QuaProgram.SetDcOffsetStatement, _Mapping]] = ..., advanceInputStream: _Optional[_Union[QuaProgram.AdvanceInputStreamStatement, _Mapping]] = ..., strictTiming: _Optional[_Union[QuaProgram.StrictTimingStatement, _Mapping]] = ..., fastFrameRotation: _Optional[_Union[QuaProgram.FastFrameRotationStatement, _Mapping]] = ..., updateOscillatorFrequency: _Optional[_Union[QuaProgram.UpdateOscillatorFrequencyStatement, _Mapping]] = ..., resetGlobalPhase: _Optional[_Union[QuaProgram.ResetGlobalPhaseStatement, _Mapping]] = ..., loadWaveform: _Optional[_Union[QuaProgram.LoadWaveformStatement, _Mapping]] = ..., sendToExternalStream: _Optional[_Union[QuaProgram.SendToExternalStreamStatement, _Mapping]] = ..., receiveFromExternalStream: _Optional[_Union[QuaProgram.ReceiveFromExternalStreamStatement, _Mapping]] = ..., globalVariableAssignment: _Optional[_Union[QuaProgram.GlobalVariableAssignmentStatement, _Mapping]] = ..., arbitrary: _Optional[_Union[QuaProgram.ArbitraryStatement, _Mapping]] = ..., arbitraryContext: _Optional[_Union[QuaProgram.ArbitraryContextStatement, _Mapping]] = ..., **kwargs) -> None: ...
    class AmpMultiplier(_message.Message):
        __slots__ = ("loc", "v0", "v1", "v2", "v3")
        LOC_FIELD_NUMBER: _ClassVar[int]
        V0_FIELD_NUMBER: _ClassVar[int]
        V1_FIELD_NUMBER: _ClassVar[int]
        V2_FIELD_NUMBER: _ClassVar[int]
        V3_FIELD_NUMBER: _ClassVar[int]
        loc: str
        v0: QuaProgram.AnyScalarExpression
        v1: QuaProgram.AnyScalarExpression
        v2: QuaProgram.AnyScalarExpression
        v3: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., v0: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., v1: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., v2: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., v3: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class RampToZeroStatement(_message.Message):
        __slots__ = ("loc", "qe", "duration")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        DURATION_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        duration: _wrappers_pb2.UInt32Value
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., duration: _Optional[_Union[_wrappers_pb2.UInt32Value, _Mapping]] = ...) -> None: ...
    class RampPulse(_message.Message):
        __slots__ = ("loc", "value")
        LOC_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        loc: str
        value: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., value: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class Chirp(_message.Message):
        __slots__ = ("loc", "scalarRate", "arrayRate", "units", "times", "continueChirp")
        class Units(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            HzPerNanoSec: _ClassVar[QuaProgram.Chirp.Units]
            mHzPerNanoSec: _ClassVar[QuaProgram.Chirp.Units]
            uHzPerNanoSec: _ClassVar[QuaProgram.Chirp.Units]
            nHzPerNanoSec: _ClassVar[QuaProgram.Chirp.Units]
            pHzPerNanoSec: _ClassVar[QuaProgram.Chirp.Units]
        HzPerNanoSec: QuaProgram.Chirp.Units
        mHzPerNanoSec: QuaProgram.Chirp.Units
        uHzPerNanoSec: QuaProgram.Chirp.Units
        nHzPerNanoSec: QuaProgram.Chirp.Units
        pHzPerNanoSec: QuaProgram.Chirp.Units
        LOC_FIELD_NUMBER: _ClassVar[int]
        SCALARRATE_FIELD_NUMBER: _ClassVar[int]
        ARRAYRATE_FIELD_NUMBER: _ClassVar[int]
        UNITS_FIELD_NUMBER: _ClassVar[int]
        TIMES_FIELD_NUMBER: _ClassVar[int]
        CONTINUECHIRP_FIELD_NUMBER: _ClassVar[int]
        loc: str
        scalarRate: QuaProgram.AnyScalarExpression
        arrayRate: QuaProgram.ArrayVarRefExpression
        units: QuaProgram.Chirp.Units
        times: _containers.RepeatedScalarFieldContainer[int]
        continueChirp: bool
        def __init__(self, loc: _Optional[str] = ..., scalarRate: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., arrayRate: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., units: _Optional[_Union[QuaProgram.Chirp.Units, str]] = ..., times: _Optional[_Iterable[int]] = ..., continueChirp: bool = ...) -> None: ...
    class PlayStatement(_message.Message):
        __slots__ = ("loc", "qe", "pulse", "namedPulse", "rampPulse", "amp", "duration", "condition", "port_condition", "targetInput", "chirp", "truncate", "timestampLabel")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        PULSE_FIELD_NUMBER: _ClassVar[int]
        NAMEDPULSE_FIELD_NUMBER: _ClassVar[int]
        RAMPPULSE_FIELD_NUMBER: _ClassVar[int]
        AMP_FIELD_NUMBER: _ClassVar[int]
        DURATION_FIELD_NUMBER: _ClassVar[int]
        CONDITION_FIELD_NUMBER: _ClassVar[int]
        PORT_CONDITION_FIELD_NUMBER: _ClassVar[int]
        TARGETINPUT_FIELD_NUMBER: _ClassVar[int]
        CHIRP_FIELD_NUMBER: _ClassVar[int]
        TRUNCATE_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMPLABEL_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        pulse: QuaProgram.PulseReference
        namedPulse: QuaProgram.PulseReference
        rampPulse: QuaProgram.RampPulse
        amp: QuaProgram.AmpMultiplier
        duration: QuaProgram.AnyScalarExpression
        condition: QuaProgram.AnyScalarExpression
        port_condition: QuaProgram.AnyScalarExpression
        targetInput: str
        chirp: QuaProgram.Chirp
        truncate: QuaProgram.AnyScalarExpression
        timestampLabel: str
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., pulse: _Optional[_Union[QuaProgram.PulseReference, _Mapping]] = ..., namedPulse: _Optional[_Union[QuaProgram.PulseReference, _Mapping]] = ..., rampPulse: _Optional[_Union[QuaProgram.RampPulse, _Mapping]] = ..., amp: _Optional[_Union[QuaProgram.AmpMultiplier, _Mapping]] = ..., duration: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., condition: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., port_condition: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., targetInput: _Optional[str] = ..., chirp: _Optional[_Union[QuaProgram.Chirp, _Mapping]] = ..., truncate: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., timestampLabel: _Optional[str] = ...) -> None: ...
    class UpdateFrequencyStatement(_message.Message):
        __slots__ = ("loc", "qe", "value", "units", "keepPhase")
        class Units(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            Hz: _ClassVar[QuaProgram.UpdateFrequencyStatement.Units]
            mHz: _ClassVar[QuaProgram.UpdateFrequencyStatement.Units]
            uHz: _ClassVar[QuaProgram.UpdateFrequencyStatement.Units]
            nHz: _ClassVar[QuaProgram.UpdateFrequencyStatement.Units]
            pHz: _ClassVar[QuaProgram.UpdateFrequencyStatement.Units]
        Hz: QuaProgram.UpdateFrequencyStatement.Units
        mHz: QuaProgram.UpdateFrequencyStatement.Units
        uHz: QuaProgram.UpdateFrequencyStatement.Units
        nHz: QuaProgram.UpdateFrequencyStatement.Units
        pHz: QuaProgram.UpdateFrequencyStatement.Units
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        UNITS_FIELD_NUMBER: _ClassVar[int]
        KEEPPHASE_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        value: QuaProgram.AnyScalarExpression
        units: QuaProgram.UpdateFrequencyStatement.Units
        keepPhase: bool
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., value: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., units: _Optional[_Union[QuaProgram.UpdateFrequencyStatement.Units, str]] = ..., keepPhase: bool = ...) -> None: ...
    class UpdateOscillatorFrequencyStatement(_message.Message):
        __slots__ = ("loc", "qe", "value", "units", "keepPhase", "updateComponent")
        class Units(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            GHz: _ClassVar[QuaProgram.UpdateOscillatorFrequencyStatement.Units]
            MHz: _ClassVar[QuaProgram.UpdateOscillatorFrequencyStatement.Units]
            KHz: _ClassVar[QuaProgram.UpdateOscillatorFrequencyStatement.Units]
            Hz: _ClassVar[QuaProgram.UpdateOscillatorFrequencyStatement.Units]
        GHz: QuaProgram.UpdateOscillatorFrequencyStatement.Units
        MHz: QuaProgram.UpdateOscillatorFrequencyStatement.Units
        KHz: QuaProgram.UpdateOscillatorFrequencyStatement.Units
        Hz: QuaProgram.UpdateOscillatorFrequencyStatement.Units
        class UpdateComponentSelection(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            both: _ClassVar[QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection]
            upconverter: _ClassVar[QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection]
            downconverter: _ClassVar[QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection]
        both: QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection
        upconverter: QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection
        downconverter: QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        UNITS_FIELD_NUMBER: _ClassVar[int]
        KEEPPHASE_FIELD_NUMBER: _ClassVar[int]
        UPDATECOMPONENT_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        value: QuaProgram.AnyScalarExpression
        units: QuaProgram.UpdateOscillatorFrequencyStatement.Units
        keepPhase: bool
        updateComponent: QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., value: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., units: _Optional[_Union[QuaProgram.UpdateOscillatorFrequencyStatement.Units, str]] = ..., keepPhase: bool = ..., updateComponent: _Optional[_Union[QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection, str]] = ...) -> None: ...
    class SetDcOffsetStatement(_message.Message):
        __slots__ = ("loc", "qe", "qeInputReference", "offset")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        QEINPUTREFERENCE_FIELD_NUMBER: _ClassVar[int]
        OFFSET_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        qeInputReference: str
        offset: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., qeInputReference: _Optional[str] = ..., offset: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class ZRotationStatement(_message.Message):
        __slots__ = ("loc", "qe", "value")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        value: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., value: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class ResetFrameStatement(_message.Message):
        __slots__ = ("loc", "qe")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ...) -> None: ...
    class FastFrameRotationStatement(_message.Message):
        __slots__ = ("loc", "qe", "cosine", "sine")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        COSINE_FIELD_NUMBER: _ClassVar[int]
        SINE_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        cosine: QuaProgram.AnyScalarExpression
        sine: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., cosine: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., sine: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class AlignStatement(_message.Message):
        __slots__ = ("loc", "qe")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: _containers.RepeatedCompositeFieldContainer[QuaProgram.QuantumElementReference]
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Iterable[_Union[QuaProgram.QuantumElementReference, _Mapping]]] = ...) -> None: ...
    class MeasureProcess(_message.Message):
        __slots__ = ("analog", "digital")
        ANALOG_FIELD_NUMBER: _ClassVar[int]
        DIGITAL_FIELD_NUMBER: _ClassVar[int]
        analog: QuaProgram.AnalogMeasureProcess
        digital: QuaProgram.DigitalMeasureProcess
        def __init__(self, analog: _Optional[_Union[QuaProgram.AnalogMeasureProcess, _Mapping]] = ..., digital: _Optional[_Union[QuaProgram.DigitalMeasureProcess, _Mapping]] = ...) -> None: ...
    class AnalogMeasureProcess(_message.Message):
        __slots__ = ("loc", "bareIntegration", "demodIntegration", "rawTimeTagging", "dualBareIntegration", "dualDemodIntegration", "highResTimeTagging")
        class BareIntegration(_message.Message):
            __slots__ = ("integration", "target", "elementOutput")
            INTEGRATION_FIELD_NUMBER: _ClassVar[int]
            TARGET_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT_FIELD_NUMBER: _ClassVar[int]
            integration: QuaProgram.IntegrationWeightReference
            target: QuaProgram.AnalogProcessTarget
            elementOutput: str
            def __init__(self, integration: _Optional[_Union[QuaProgram.IntegrationWeightReference, _Mapping]] = ..., target: _Optional[_Union[QuaProgram.AnalogProcessTarget, _Mapping]] = ..., elementOutput: _Optional[str] = ...) -> None: ...
        class DualBareIntegration(_message.Message):
            __slots__ = ("integration1", "integration2", "target", "elementOutput1", "elementOutput2")
            INTEGRATION1_FIELD_NUMBER: _ClassVar[int]
            INTEGRATION2_FIELD_NUMBER: _ClassVar[int]
            TARGET_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT1_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT2_FIELD_NUMBER: _ClassVar[int]
            integration1: QuaProgram.IntegrationWeightReference
            integration2: QuaProgram.IntegrationWeightReference
            target: QuaProgram.AnalogProcessTarget
            elementOutput1: str
            elementOutput2: str
            def __init__(self, integration1: _Optional[_Union[QuaProgram.IntegrationWeightReference, _Mapping]] = ..., integration2: _Optional[_Union[QuaProgram.IntegrationWeightReference, _Mapping]] = ..., target: _Optional[_Union[QuaProgram.AnalogProcessTarget, _Mapping]] = ..., elementOutput1: _Optional[str] = ..., elementOutput2: _Optional[str] = ...) -> None: ...
        class DemodIntegration(_message.Message):
            __slots__ = ("integration", "target", "elementOutput")
            INTEGRATION_FIELD_NUMBER: _ClassVar[int]
            TARGET_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT_FIELD_NUMBER: _ClassVar[int]
            integration: QuaProgram.IntegrationWeightReference
            target: QuaProgram.AnalogProcessTarget
            elementOutput: str
            def __init__(self, integration: _Optional[_Union[QuaProgram.IntegrationWeightReference, _Mapping]] = ..., target: _Optional[_Union[QuaProgram.AnalogProcessTarget, _Mapping]] = ..., elementOutput: _Optional[str] = ...) -> None: ...
        class DualDemodIntegration(_message.Message):
            __slots__ = ("integration1", "integration2", "target", "elementOutput1", "elementOutput2")
            INTEGRATION1_FIELD_NUMBER: _ClassVar[int]
            INTEGRATION2_FIELD_NUMBER: _ClassVar[int]
            TARGET_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT1_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT2_FIELD_NUMBER: _ClassVar[int]
            integration1: QuaProgram.IntegrationWeightReference
            integration2: QuaProgram.IntegrationWeightReference
            target: QuaProgram.AnalogProcessTarget
            elementOutput1: str
            elementOutput2: str
            def __init__(self, integration1: _Optional[_Union[QuaProgram.IntegrationWeightReference, _Mapping]] = ..., integration2: _Optional[_Union[QuaProgram.IntegrationWeightReference, _Mapping]] = ..., target: _Optional[_Union[QuaProgram.AnalogProcessTarget, _Mapping]] = ..., elementOutput1: _Optional[str] = ..., elementOutput2: _Optional[str] = ...) -> None: ...
        class RawTimeTagging(_message.Message):
            __slots__ = ("target", "targetLen", "maxTime", "elementOutput")
            TARGET_FIELD_NUMBER: _ClassVar[int]
            TARGETLEN_FIELD_NUMBER: _ClassVar[int]
            MAXTIME_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT_FIELD_NUMBER: _ClassVar[int]
            target: QuaProgram.ArrayVarRefExpression
            targetLen: QuaProgram.VarRefExpression
            maxTime: int
            elementOutput: str
            def __init__(self, target: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., targetLen: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., maxTime: _Optional[int] = ..., elementOutput: _Optional[str] = ...) -> None: ...
        class HighResTimeTagging(_message.Message):
            __slots__ = ("target", "targetLen", "maxTime", "elementOutput")
            TARGET_FIELD_NUMBER: _ClassVar[int]
            TARGETLEN_FIELD_NUMBER: _ClassVar[int]
            MAXTIME_FIELD_NUMBER: _ClassVar[int]
            ELEMENTOUTPUT_FIELD_NUMBER: _ClassVar[int]
            target: QuaProgram.ArrayVarRefExpression
            targetLen: QuaProgram.VarRefExpression
            maxTime: int
            elementOutput: str
            def __init__(self, target: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., targetLen: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., maxTime: _Optional[int] = ..., elementOutput: _Optional[str] = ...) -> None: ...
        LOC_FIELD_NUMBER: _ClassVar[int]
        BAREINTEGRATION_FIELD_NUMBER: _ClassVar[int]
        DEMODINTEGRATION_FIELD_NUMBER: _ClassVar[int]
        RAWTIMETAGGING_FIELD_NUMBER: _ClassVar[int]
        DUALBAREINTEGRATION_FIELD_NUMBER: _ClassVar[int]
        DUALDEMODINTEGRATION_FIELD_NUMBER: _ClassVar[int]
        HIGHRESTIMETAGGING_FIELD_NUMBER: _ClassVar[int]
        loc: str
        bareIntegration: QuaProgram.AnalogMeasureProcess.BareIntegration
        demodIntegration: QuaProgram.AnalogMeasureProcess.DemodIntegration
        rawTimeTagging: QuaProgram.AnalogMeasureProcess.RawTimeTagging
        dualBareIntegration: QuaProgram.AnalogMeasureProcess.DualBareIntegration
        dualDemodIntegration: QuaProgram.AnalogMeasureProcess.DualDemodIntegration
        highResTimeTagging: QuaProgram.AnalogMeasureProcess.HighResTimeTagging
        def __init__(self, loc: _Optional[str] = ..., bareIntegration: _Optional[_Union[QuaProgram.AnalogMeasureProcess.BareIntegration, _Mapping]] = ..., demodIntegration: _Optional[_Union[QuaProgram.AnalogMeasureProcess.DemodIntegration, _Mapping]] = ..., rawTimeTagging: _Optional[_Union[QuaProgram.AnalogMeasureProcess.RawTimeTagging, _Mapping]] = ..., dualBareIntegration: _Optional[_Union[QuaProgram.AnalogMeasureProcess.DualBareIntegration, _Mapping]] = ..., dualDemodIntegration: _Optional[_Union[QuaProgram.AnalogMeasureProcess.DualDemodIntegration, _Mapping]] = ..., highResTimeTagging: _Optional[_Union[QuaProgram.AnalogMeasureProcess.HighResTimeTagging, _Mapping]] = ...) -> None: ...
    class DigitalMeasureProcess(_message.Message):
        __slots__ = ("loc", "rawTimeTagging", "counting")
        class RawTimeTagging(_message.Message):
            __slots__ = ("elementOutput", "target", "targetLen", "maxTime")
            ELEMENTOUTPUT_FIELD_NUMBER: _ClassVar[int]
            TARGET_FIELD_NUMBER: _ClassVar[int]
            TARGETLEN_FIELD_NUMBER: _ClassVar[int]
            MAXTIME_FIELD_NUMBER: _ClassVar[int]
            elementOutput: str
            target: QuaProgram.ArrayVarRefExpression
            targetLen: QuaProgram.VarRefExpression
            maxTime: int
            def __init__(self, elementOutput: _Optional[str] = ..., target: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., targetLen: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., maxTime: _Optional[int] = ...) -> None: ...
        class Counting(_message.Message):
            __slots__ = ("elementOutputs", "target", "maxTime")
            ELEMENTOUTPUTS_FIELD_NUMBER: _ClassVar[int]
            TARGET_FIELD_NUMBER: _ClassVar[int]
            MAXTIME_FIELD_NUMBER: _ClassVar[int]
            elementOutputs: _containers.RepeatedScalarFieldContainer[str]
            target: QuaProgram.VarRefExpression
            maxTime: int
            def __init__(self, elementOutputs: _Optional[_Iterable[str]] = ..., target: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., maxTime: _Optional[int] = ...) -> None: ...
        LOC_FIELD_NUMBER: _ClassVar[int]
        RAWTIMETAGGING_FIELD_NUMBER: _ClassVar[int]
        COUNTING_FIELD_NUMBER: _ClassVar[int]
        loc: str
        rawTimeTagging: QuaProgram.DigitalMeasureProcess.RawTimeTagging
        counting: QuaProgram.DigitalMeasureProcess.Counting
        def __init__(self, loc: _Optional[str] = ..., rawTimeTagging: _Optional[_Union[QuaProgram.DigitalMeasureProcess.RawTimeTagging, _Mapping]] = ..., counting: _Optional[_Union[QuaProgram.DigitalMeasureProcess.Counting, _Mapping]] = ...) -> None: ...
    class AnalogProcessTarget(_message.Message):
        __slots__ = ("loc", "scalarProcess", "vectorProcess")
        class ScalarProcessTarget(_message.Message):
            __slots__ = ("variable", "arrayCell")
            VARIABLE_FIELD_NUMBER: _ClassVar[int]
            ARRAYCELL_FIELD_NUMBER: _ClassVar[int]
            variable: QuaProgram.VarRefExpression
            arrayCell: QuaProgram.ArrayCellRefExpression
            def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., arrayCell: _Optional[_Union[QuaProgram.ArrayCellRefExpression, _Mapping]] = ...) -> None: ...
        class VectorProcessTarget(_message.Message):
            __slots__ = ("array", "timeDivision")
            ARRAY_FIELD_NUMBER: _ClassVar[int]
            TIMEDIVISION_FIELD_NUMBER: _ClassVar[int]
            array: QuaProgram.ArrayVarRefExpression
            timeDivision: QuaProgram.AnalogProcessTarget.TimeDivision
            def __init__(self, array: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., timeDivision: _Optional[_Union[QuaProgram.AnalogProcessTarget.TimeDivision, _Mapping]] = ...) -> None: ...
        class TimeDivision(_message.Message):
            __slots__ = ("sliced", "accumulated", "movingWindow")
            SLICED_FIELD_NUMBER: _ClassVar[int]
            ACCUMULATED_FIELD_NUMBER: _ClassVar[int]
            MOVINGWINDOW_FIELD_NUMBER: _ClassVar[int]
            sliced: QuaProgram.AnalogTimeDivision.Sliced
            accumulated: QuaProgram.AnalogTimeDivision.Accumulated
            movingWindow: QuaProgram.AnalogTimeDivision.MovingWindow
            def __init__(self, sliced: _Optional[_Union[QuaProgram.AnalogTimeDivision.Sliced, _Mapping]] = ..., accumulated: _Optional[_Union[QuaProgram.AnalogTimeDivision.Accumulated, _Mapping]] = ..., movingWindow: _Optional[_Union[QuaProgram.AnalogTimeDivision.MovingWindow, _Mapping]] = ...) -> None: ...
        LOC_FIELD_NUMBER: _ClassVar[int]
        SCALARPROCESS_FIELD_NUMBER: _ClassVar[int]
        VECTORPROCESS_FIELD_NUMBER: _ClassVar[int]
        loc: str
        scalarProcess: QuaProgram.AnalogProcessTarget.ScalarProcessTarget
        vectorProcess: QuaProgram.AnalogProcessTarget.VectorProcessTarget
        def __init__(self, loc: _Optional[str] = ..., scalarProcess: _Optional[_Union[QuaProgram.AnalogProcessTarget.ScalarProcessTarget, _Mapping]] = ..., vectorProcess: _Optional[_Union[QuaProgram.AnalogProcessTarget.VectorProcessTarget, _Mapping]] = ...) -> None: ...
    class AnalogTimeDivision(_message.Message):
        __slots__ = ()
        class Sliced(_message.Message):
            __slots__ = ("loc", "samplesPerChunk")
            LOC_FIELD_NUMBER: _ClassVar[int]
            SAMPLESPERCHUNK_FIELD_NUMBER: _ClassVar[int]
            loc: str
            samplesPerChunk: int
            def __init__(self, loc: _Optional[str] = ..., samplesPerChunk: _Optional[int] = ...) -> None: ...
        class Accumulated(_message.Message):
            __slots__ = ("loc", "samplesPerChunk")
            LOC_FIELD_NUMBER: _ClassVar[int]
            SAMPLESPERCHUNK_FIELD_NUMBER: _ClassVar[int]
            loc: str
            samplesPerChunk: int
            def __init__(self, loc: _Optional[str] = ..., samplesPerChunk: _Optional[int] = ...) -> None: ...
        class MovingWindow(_message.Message):
            __slots__ = ("loc", "samplesPerChunk", "chunksPerWindow")
            LOC_FIELD_NUMBER: _ClassVar[int]
            SAMPLESPERCHUNK_FIELD_NUMBER: _ClassVar[int]
            CHUNKSPERWINDOW_FIELD_NUMBER: _ClassVar[int]
            loc: str
            samplesPerChunk: int
            chunksPerWindow: int
            def __init__(self, loc: _Optional[str] = ..., samplesPerChunk: _Optional[int] = ..., chunksPerWindow: _Optional[int] = ...) -> None: ...
        def __init__(self) -> None: ...
    class MeasureOutput(_message.Message):
        __slots__ = ("loc", "integration", "variable", "output", "target")
        class Target(_message.Message):
            __slots__ = ("variable", "arrayCell")
            VARIABLE_FIELD_NUMBER: _ClassVar[int]
            ARRAYCELL_FIELD_NUMBER: _ClassVar[int]
            variable: QuaProgram.VarRefExpression
            arrayCell: QuaProgram.ArrayCellRefExpression
            def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., arrayCell: _Optional[_Union[QuaProgram.ArrayCellRefExpression, _Mapping]] = ...) -> None: ...
        LOC_FIELD_NUMBER: _ClassVar[int]
        INTEGRATION_FIELD_NUMBER: _ClassVar[int]
        VARIABLE_FIELD_NUMBER: _ClassVar[int]
        OUTPUT_FIELD_NUMBER: _ClassVar[int]
        TARGET_FIELD_NUMBER: _ClassVar[int]
        loc: str
        integration: QuaProgram.IntegrationWeightReference
        variable: QuaProgram.VarRefExpression
        output: str
        target: QuaProgram.MeasureOutput.Target
        def __init__(self, loc: _Optional[str] = ..., integration: _Optional[_Union[QuaProgram.IntegrationWeightReference, _Mapping]] = ..., variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., output: _Optional[str] = ..., target: _Optional[_Union[QuaProgram.MeasureOutput.Target, _Mapping]] = ...) -> None: ...
    class MeasureStatement(_message.Message):
        __slots__ = ("loc", "qe", "pulse", "amp", "outputs", "streamAs", "measureProcesses", "timestampLabel")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        PULSE_FIELD_NUMBER: _ClassVar[int]
        AMP_FIELD_NUMBER: _ClassVar[int]
        OUTPUTS_FIELD_NUMBER: _ClassVar[int]
        STREAMAS_FIELD_NUMBER: _ClassVar[int]
        MEASUREPROCESSES_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMPLABEL_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        pulse: QuaProgram.PulseReference
        amp: QuaProgram.AmpMultiplier
        outputs: _containers.RepeatedCompositeFieldContainer[QuaProgram.MeasureOutput]
        streamAs: str
        measureProcesses: _containers.RepeatedCompositeFieldContainer[QuaProgram.MeasureProcess]
        timestampLabel: str
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., pulse: _Optional[_Union[QuaProgram.PulseReference, _Mapping]] = ..., amp: _Optional[_Union[QuaProgram.AmpMultiplier, _Mapping]] = ..., outputs: _Optional[_Iterable[_Union[QuaProgram.MeasureOutput, _Mapping]]] = ..., streamAs: _Optional[str] = ..., measureProcesses: _Optional[_Iterable[_Union[QuaProgram.MeasureProcess, _Mapping]]] = ..., timestampLabel: _Optional[str] = ...) -> None: ...
    class WaitStatement(_message.Message):
        __slots__ = ("loc", "qe", "time")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        TIME_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: _containers.RepeatedCompositeFieldContainer[QuaProgram.QuantumElementReference]
        time: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Iterable[_Union[QuaProgram.QuantumElementReference, _Mapping]]] = ..., time: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class ArbitraryStatement(_message.Message):
        __slots__ = ("loc", "name", "data")
        LOC_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        DATA_FIELD_NUMBER: _ClassVar[int]
        loc: str
        name: str
        data: _any_pb2.Any
        def __init__(self, loc: _Optional[str] = ..., name: _Optional[str] = ..., data: _Optional[_Union[_any_pb2.Any, _Mapping]] = ...) -> None: ...
    class ArbitraryContextStatement(_message.Message):
        __slots__ = ("loc", "name", "data", "body")
        LOC_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        DATA_FIELD_NUMBER: _ClassVar[int]
        BODY_FIELD_NUMBER: _ClassVar[int]
        loc: str
        name: str
        data: _any_pb2.Any
        body: QuaProgram.StatementsCollection
        def __init__(self, loc: _Optional[str] = ..., name: _Optional[str] = ..., data: _Optional[_Union[_any_pb2.Any, _Mapping]] = ..., body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ...) -> None: ...
    class WaitForTriggerStatement(_message.Message):
        __slots__ = ("loc", "qe", "pulseToPlay", "globalTrigger", "elementOutput", "timeTagTarget")
        class ElementOutput(_message.Message):
            __slots__ = ("element", "output")
            ELEMENT_FIELD_NUMBER: _ClassVar[int]
            OUTPUT_FIELD_NUMBER: _ClassVar[int]
            element: str
            output: str
            def __init__(self, element: _Optional[str] = ..., output: _Optional[str] = ...) -> None: ...
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        PULSETOPLAY_FIELD_NUMBER: _ClassVar[int]
        GLOBALTRIGGER_FIELD_NUMBER: _ClassVar[int]
        ELEMENTOUTPUT_FIELD_NUMBER: _ClassVar[int]
        TIMETAGTARGET_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: _containers.RepeatedCompositeFieldContainer[QuaProgram.QuantumElementReference]
        pulseToPlay: QuaProgram.PulseReference
        globalTrigger: _empty_pb2.Empty
        elementOutput: QuaProgram.WaitForTriggerStatement.ElementOutput
        timeTagTarget: QuaProgram.VarRefExpression
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Iterable[_Union[QuaProgram.QuantumElementReference, _Mapping]]] = ..., pulseToPlay: _Optional[_Union[QuaProgram.PulseReference, _Mapping]] = ..., globalTrigger: _Optional[_Union[_empty_pb2.Empty, _Mapping]] = ..., elementOutput: _Optional[_Union[QuaProgram.WaitForTriggerStatement.ElementOutput, _Mapping]] = ..., timeTagTarget: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ...) -> None: ...
    class UpdateCorrectionStatement(_message.Message):
        __slots__ = ("loc", "qe", "correction")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        CORRECTION_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        correction: QuaProgram.Correction
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., correction: _Optional[_Union[QuaProgram.Correction, _Mapping]] = ...) -> None: ...
    class ResetPhaseStatement(_message.Message):
        __slots__ = ("loc", "qe")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ...) -> None: ...
    class ResetGlobalPhaseStatement(_message.Message):
        __slots__ = ("loc",)
        LOC_FIELD_NUMBER: _ClassVar[int]
        loc: str
        def __init__(self, loc: _Optional[str] = ...) -> None: ...
    class SendToExternalStreamStatement(_message.Message):
        __slots__ = ("loc", "stream", "struct")
        LOC_FIELD_NUMBER: _ClassVar[int]
        STREAM_FIELD_NUMBER: _ClassVar[int]
        STRUCT_FIELD_NUMBER: _ClassVar[int]
        loc: str
        stream: QuaProgram.ExternalStreamRefExpression
        struct: QuaProgram.StructVarRefExpression
        def __init__(self, loc: _Optional[str] = ..., stream: _Optional[_Union[QuaProgram.ExternalStreamRefExpression, _Mapping]] = ..., struct: _Optional[_Union[QuaProgram.StructVarRefExpression, _Mapping]] = ...) -> None: ...
    class ReceiveFromExternalStreamStatement(_message.Message):
        __slots__ = ("loc", "stream", "struct")
        LOC_FIELD_NUMBER: _ClassVar[int]
        STREAM_FIELD_NUMBER: _ClassVar[int]
        STRUCT_FIELD_NUMBER: _ClassVar[int]
        loc: str
        stream: QuaProgram.ExternalStreamRefExpression
        struct: QuaProgram.StructVarRefExpression
        def __init__(self, loc: _Optional[str] = ..., stream: _Optional[_Union[QuaProgram.ExternalStreamRefExpression, _Mapping]] = ..., struct: _Optional[_Union[QuaProgram.StructVarRefExpression, _Mapping]] = ...) -> None: ...
    class GlobalVariableAssignmentStatement(_message.Message):
        __slots__ = ("loc", "variables")
        LOC_FIELD_NUMBER: _ClassVar[int]
        VARIABLES_FIELD_NUMBER: _ClassVar[int]
        loc: str
        variables: _containers.RepeatedCompositeFieldContainer[QuaProgram.VarRefExpression]
        def __init__(self, loc: _Optional[str] = ..., variables: _Optional[_Iterable[_Union[QuaProgram.VarRefExpression, _Mapping]]] = ...) -> None: ...
    class Correction(_message.Message):
        __slots__ = ("loc", "c0", "c1", "c2", "c3")
        LOC_FIELD_NUMBER: _ClassVar[int]
        C0_FIELD_NUMBER: _ClassVar[int]
        C1_FIELD_NUMBER: _ClassVar[int]
        C2_FIELD_NUMBER: _ClassVar[int]
        C3_FIELD_NUMBER: _ClassVar[int]
        loc: str
        c0: QuaProgram.AnyScalarExpression
        c1: QuaProgram.AnyScalarExpression
        c2: QuaProgram.AnyScalarExpression
        c3: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., c0: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., c1: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., c2: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., c3: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class SyncStatement(_message.Message):
        __slots__ = ("loc",)
        LOC_FIELD_NUMBER: _ClassVar[int]
        loc: str
        def __init__(self, loc: _Optional[str] = ...) -> None: ...
    class ElseIf(_message.Message):
        __slots__ = ("condition", "body", "loc")
        CONDITION_FIELD_NUMBER: _ClassVar[int]
        BODY_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        condition: QuaProgram.AnyScalarExpression
        body: QuaProgram.StatementsCollection
        loc: str
        def __init__(self, condition: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class IfStatement(_message.Message):
        __slots__ = ("condition", "body", "elseifs", "unsafe", "loc")
        CONDITION_FIELD_NUMBER: _ClassVar[int]
        BODY_FIELD_NUMBER: _ClassVar[int]
        ELSE_FIELD_NUMBER: _ClassVar[int]
        ELSEIFS_FIELD_NUMBER: _ClassVar[int]
        UNSAFE_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        condition: QuaProgram.AnyScalarExpression
        body: QuaProgram.StatementsCollection
        elseifs: _containers.RepeatedCompositeFieldContainer[QuaProgram.ElseIf]
        unsafe: bool
        loc: str
        def __init__(self, condition: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., elseifs: _Optional[_Iterable[_Union[QuaProgram.ElseIf, _Mapping]]] = ..., unsafe: bool = ..., loc: _Optional[str] = ..., **kwargs) -> None: ...
    class ForStatement(_message.Message):
        __slots__ = ("init", "condition", "update", "body", "loc")
        INIT_FIELD_NUMBER: _ClassVar[int]
        CONDITION_FIELD_NUMBER: _ClassVar[int]
        UPDATE_FIELD_NUMBER: _ClassVar[int]
        BODY_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        init: QuaProgram.StatementsCollection
        condition: QuaProgram.AnyScalarExpression
        update: QuaProgram.StatementsCollection
        body: QuaProgram.StatementsCollection
        loc: str
        def __init__(self, init: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., condition: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., update: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class QRunStatement(_message.Message):
        __slots__ = ("body", "loc")
        BODY_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        body: QuaProgram.StatementsCollection
        loc: str
        def __init__(self, body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class StrictTimingStatement(_message.Message):
        __slots__ = ("body", "loc")
        BODY_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        body: QuaProgram.StatementsCollection
        loc: str
        def __init__(self, body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class ForEachStatement(_message.Message):
        __slots__ = ("iterator", "body", "loc")
        class VariableWithValues(_message.Message):
            __slots__ = ("variable", "array")
            VARIABLE_FIELD_NUMBER: _ClassVar[int]
            ARRAY_FIELD_NUMBER: _ClassVar[int]
            variable: QuaProgram.VarRefExpression
            array: QuaProgram.ArrayVarRefExpression
            def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., array: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ...) -> None: ...
        ITERATOR_FIELD_NUMBER: _ClassVar[int]
        BODY_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        iterator: _containers.RepeatedCompositeFieldContainer[QuaProgram.ForEachStatement.VariableWithValues]
        body: QuaProgram.StatementsCollection
        loc: str
        def __init__(self, iterator: _Optional[_Iterable[_Union[QuaProgram.ForEachStatement.VariableWithValues, _Mapping]]] = ..., body: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class RusStatement(_message.Message):
        __slots__ = ("doBlock", "failBlock", "until", "loc")
        DOBLOCK_FIELD_NUMBER: _ClassVar[int]
        FAILBLOCK_FIELD_NUMBER: _ClassVar[int]
        UNTIL_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        doBlock: QuaProgram.StatementsCollection
        failBlock: QuaProgram.StatementsCollection
        until: QuaProgram.AnyScalarExpression
        loc: str
        def __init__(self, doBlock: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., failBlock: _Optional[_Union[QuaProgram.StatementsCollection, _Mapping]] = ..., until: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class AssignmentStatement(_message.Message):
        __slots__ = ("variable", "expression", "save", "target", "loc")
        class Target(_message.Message):
            __slots__ = ("variable", "arrayCell")
            VARIABLE_FIELD_NUMBER: _ClassVar[int]
            ARRAYCELL_FIELD_NUMBER: _ClassVar[int]
            variable: QuaProgram.VarRefExpression
            arrayCell: QuaProgram.ArrayCellRefExpression
            def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., arrayCell: _Optional[_Union[QuaProgram.ArrayCellRefExpression, _Mapping]] = ...) -> None: ...
        VARIABLE_FIELD_NUMBER: _ClassVar[int]
        EXPRESSION_FIELD_NUMBER: _ClassVar[int]
        SAVE_FIELD_NUMBER: _ClassVar[int]
        TARGET_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        variable: QuaProgram.VarRefExpression
        expression: QuaProgram.AnyScalarExpression
        save: bool
        target: QuaProgram.AssignmentStatement.Target
        loc: str
        def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., expression: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., save: bool = ..., target: _Optional[_Union[QuaProgram.AssignmentStatement.Target, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class PauseStatement(_message.Message):
        __slots__ = ("qes", "loc")
        QES_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        qes: _containers.RepeatedCompositeFieldContainer[QuaProgram.QuantumElementReference]
        loc: str
        def __init__(self, qes: _Optional[_Iterable[_Union[QuaProgram.QuantumElementReference, _Mapping]]] = ..., loc: _Optional[str] = ...) -> None: ...
    class SaveStatement(_message.Message):
        __slots__ = ("variable", "tag", "source", "loc")
        class Source(_message.Message):
            __slots__ = ("variable", "arrayCell", "literal")
            VARIABLE_FIELD_NUMBER: _ClassVar[int]
            ARRAYCELL_FIELD_NUMBER: _ClassVar[int]
            LITERAL_FIELD_NUMBER: _ClassVar[int]
            variable: QuaProgram.VarRefExpression
            arrayCell: QuaProgram.ArrayCellRefExpression
            literal: QuaProgram.LiteralExpression
            def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., arrayCell: _Optional[_Union[QuaProgram.ArrayCellRefExpression, _Mapping]] = ..., literal: _Optional[_Union[QuaProgram.LiteralExpression, _Mapping]] = ...) -> None: ...
        VARIABLE_FIELD_NUMBER: _ClassVar[int]
        TAG_FIELD_NUMBER: _ClassVar[int]
        SOURCE_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        variable: QuaProgram.VarRefExpression
        tag: str
        source: QuaProgram.SaveStatement.Source
        loc: str
        def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., tag: _Optional[str] = ..., source: _Optional[_Union[QuaProgram.SaveStatement.Source, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class LoadWaveformStatement(_message.Message):
        __slots__ = ("loc", "qe", "pulse", "waveform_index")
        LOC_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        PULSE_FIELD_NUMBER: _ClassVar[int]
        WAVEFORM_INDEX_FIELD_NUMBER: _ClassVar[int]
        loc: str
        qe: QuaProgram.QuantumElementReference
        pulse: QuaProgram.PulseReference
        waveform_index: QuaProgram.AnyScalarExpression
        def __init__(self, loc: _Optional[str] = ..., qe: _Optional[_Union[QuaProgram.QuantumElementReference, _Mapping]] = ..., pulse: _Optional[_Union[QuaProgram.PulseReference, _Mapping]] = ..., waveform_index: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ...) -> None: ...
    class StatementsCollection(_message.Message):
        __slots__ = ("statements",)
        STATEMENTS_FIELD_NUMBER: _ClassVar[int]
        statements: _containers.RepeatedCompositeFieldContainer[QuaProgram.AnyStatement]
        def __init__(self, statements: _Optional[_Iterable[_Union[QuaProgram.AnyStatement, _Mapping]]] = ...) -> None: ...
    class AnyScalarExpression(_message.Message):
        __slots__ = ("variable", "literal", "binaryOperation", "arrayCell", "arrayLength", "libFunction", "function", "broadcast", "globalVariable")
        VARIABLE_FIELD_NUMBER: _ClassVar[int]
        LITERAL_FIELD_NUMBER: _ClassVar[int]
        BINARYOPERATION_FIELD_NUMBER: _ClassVar[int]
        ARRAYCELL_FIELD_NUMBER: _ClassVar[int]
        ARRAYLENGTH_FIELD_NUMBER: _ClassVar[int]
        LIBFUNCTION_FIELD_NUMBER: _ClassVar[int]
        FUNCTION_FIELD_NUMBER: _ClassVar[int]
        BROADCAST_FIELD_NUMBER: _ClassVar[int]
        GLOBALVARIABLE_FIELD_NUMBER: _ClassVar[int]
        variable: QuaProgram.VarRefExpression
        literal: QuaProgram.LiteralExpression
        binaryOperation: QuaProgram.BinaryExpression
        arrayCell: QuaProgram.ArrayCellRefExpression
        arrayLength: QuaProgram.ArrayLengthExpression
        libFunction: QuaProgram.LibFunctionExpression
        function: QuaProgram.FunctionExpression
        broadcast: QuaProgram.BroadcastExpression
        globalVariable: QuaProgram.GlobalVarRefExpression
        def __init__(self, variable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., literal: _Optional[_Union[QuaProgram.LiteralExpression, _Mapping]] = ..., binaryOperation: _Optional[_Union[QuaProgram.BinaryExpression, _Mapping]] = ..., arrayCell: _Optional[_Union[QuaProgram.ArrayCellRefExpression, _Mapping]] = ..., arrayLength: _Optional[_Union[QuaProgram.ArrayLengthExpression, _Mapping]] = ..., libFunction: _Optional[_Union[QuaProgram.LibFunctionExpression, _Mapping]] = ..., function: _Optional[_Union[QuaProgram.FunctionExpression, _Mapping]] = ..., broadcast: _Optional[_Union[QuaProgram.BroadcastExpression, _Mapping]] = ..., globalVariable: _Optional[_Union[QuaProgram.GlobalVarRefExpression, _Mapping]] = ...) -> None: ...
    class AdvanceInputStreamStatement(_message.Message):
        __slots__ = ("streamVariable", "streamArray", "loc")
        STREAMVARIABLE_FIELD_NUMBER: _ClassVar[int]
        STREAMARRAY_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        streamVariable: QuaProgram.VarRefExpression
        streamArray: QuaProgram.ArrayVarRefExpression
        loc: str
        def __init__(self, streamVariable: _Optional[_Union[QuaProgram.VarRefExpression, _Mapping]] = ..., streamArray: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class ArrayLengthExpression(_message.Message):
        __slots__ = ("array", "loc")
        ARRAY_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        array: QuaProgram.ArrayVarRefExpression
        loc: str
        def __init__(self, array: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class VarRefExpression(_message.Message):
        __slots__ = ("name", "ioNumber", "loc")
        NAME_FIELD_NUMBER: _ClassVar[int]
        IONUMBER_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        name: str
        ioNumber: int
        loc: str
        def __init__(self, name: _Optional[str] = ..., ioNumber: _Optional[int] = ..., loc: _Optional[str] = ...) -> None: ...
    class StructVarRefExpression(_message.Message):
        __slots__ = ("name", "loc")
        NAME_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        name: str
        loc: str
        def __init__(self, name: _Optional[str] = ..., loc: _Optional[str] = ...) -> None: ...
    class ExternalStreamRefExpression(_message.Message):
        __slots__ = ("stream_id", "loc")
        STREAM_ID_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        stream_id: int
        loc: str
        def __init__(self, stream_id: _Optional[int] = ..., loc: _Optional[str] = ...) -> None: ...
    class ArrayVarRefExpression(_message.Message):
        __slots__ = ("name", "structVar", "loc")
        NAME_FIELD_NUMBER: _ClassVar[int]
        STRUCTVAR_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        name: str
        structVar: QuaProgram.StructVarRefExpression
        loc: str
        def __init__(self, name: _Optional[str] = ..., structVar: _Optional[_Union[QuaProgram.StructVarRefExpression, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class ArrayCellRefExpression(_message.Message):
        __slots__ = ("arrayVar", "index", "loc")
        ARRAYVAR_FIELD_NUMBER: _ClassVar[int]
        INDEX_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        arrayVar: QuaProgram.ArrayVarRefExpression
        index: QuaProgram.AnyScalarExpression
        loc: str
        def __init__(self, arrayVar: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ..., index: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class LiteralExpression(_message.Message):
        __slots__ = ("value", "type", "loc")
        VALUE_FIELD_NUMBER: _ClassVar[int]
        TYPE_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        value: str
        type: QuaProgram.Type
        loc: str
        def __init__(self, value: _Optional[str] = ..., type: _Optional[_Union[QuaProgram.Type, str]] = ..., loc: _Optional[str] = ...) -> None: ...
    class BinaryExpression(_message.Message):
        __slots__ = ("op", "left", "right", "loc")
        class BinaryOperator(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            ADD: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            SUB: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            MULT: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            DIV: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            AND: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            OR: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            XOR: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            LT: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            LET: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            GT: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            GET: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            EQ: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            SHL: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
            SHR: _ClassVar[QuaProgram.BinaryExpression.BinaryOperator]
        ADD: QuaProgram.BinaryExpression.BinaryOperator
        SUB: QuaProgram.BinaryExpression.BinaryOperator
        MULT: QuaProgram.BinaryExpression.BinaryOperator
        DIV: QuaProgram.BinaryExpression.BinaryOperator
        AND: QuaProgram.BinaryExpression.BinaryOperator
        OR: QuaProgram.BinaryExpression.BinaryOperator
        XOR: QuaProgram.BinaryExpression.BinaryOperator
        LT: QuaProgram.BinaryExpression.BinaryOperator
        LET: QuaProgram.BinaryExpression.BinaryOperator
        GT: QuaProgram.BinaryExpression.BinaryOperator
        GET: QuaProgram.BinaryExpression.BinaryOperator
        EQ: QuaProgram.BinaryExpression.BinaryOperator
        SHL: QuaProgram.BinaryExpression.BinaryOperator
        SHR: QuaProgram.BinaryExpression.BinaryOperator
        OP_FIELD_NUMBER: _ClassVar[int]
        LEFT_FIELD_NUMBER: _ClassVar[int]
        RIGHT_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        op: QuaProgram.BinaryExpression.BinaryOperator
        left: QuaProgram.AnyScalarExpression
        right: QuaProgram.AnyScalarExpression
        loc: str
        def __init__(self, op: _Optional[_Union[QuaProgram.BinaryExpression.BinaryOperator, str]] = ..., left: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., right: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class LibFunctionExpression(_message.Message):
        __slots__ = ("functionName", "arguments", "libraryName", "loc")
        class Argument(_message.Message):
            __slots__ = ("scalar", "array")
            SCALAR_FIELD_NUMBER: _ClassVar[int]
            ARRAY_FIELD_NUMBER: _ClassVar[int]
            scalar: QuaProgram.AnyScalarExpression
            array: QuaProgram.ArrayVarRefExpression
            def __init__(self, scalar: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., array: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ...) -> None: ...
        FUNCTIONNAME_FIELD_NUMBER: _ClassVar[int]
        ARGUMENTS_FIELD_NUMBER: _ClassVar[int]
        LIBRARYNAME_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        functionName: str
        arguments: _containers.RepeatedCompositeFieldContainer[QuaProgram.LibFunctionExpression.Argument]
        libraryName: str
        loc: str
        def __init__(self, functionName: _Optional[str] = ..., arguments: _Optional[_Iterable[_Union[QuaProgram.LibFunctionExpression.Argument, _Mapping]]] = ..., libraryName: _Optional[str] = ..., loc: _Optional[str] = ...) -> None: ...
    class FunctionExpression(_message.Message):
        __slots__ = ("xor", "loc")
        class ScalarOrVectorArgument(_message.Message):
            __slots__ = ("scalar", "array")
            SCALAR_FIELD_NUMBER: _ClassVar[int]
            ARRAY_FIELD_NUMBER: _ClassVar[int]
            scalar: QuaProgram.AnyScalarExpression
            array: QuaProgram.ArrayVarRefExpression
            def __init__(self, scalar: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., array: _Optional[_Union[QuaProgram.ArrayVarRefExpression, _Mapping]] = ...) -> None: ...
        class AndFunction(_message.Message):
            __slots__ = ("values",)
            VALUES_FIELD_NUMBER: _ClassVar[int]
            values: _containers.RepeatedCompositeFieldContainer[QuaProgram.FunctionExpression.ScalarOrVectorArgument]
            def __init__(self, values: _Optional[_Iterable[_Union[QuaProgram.FunctionExpression.ScalarOrVectorArgument, _Mapping]]] = ...) -> None: ...
        class OrFunction(_message.Message):
            __slots__ = ("values",)
            VALUES_FIELD_NUMBER: _ClassVar[int]
            values: _containers.RepeatedCompositeFieldContainer[QuaProgram.FunctionExpression.ScalarOrVectorArgument]
            def __init__(self, values: _Optional[_Iterable[_Union[QuaProgram.FunctionExpression.ScalarOrVectorArgument, _Mapping]]] = ...) -> None: ...
        class XorFunction(_message.Message):
            __slots__ = ("values",)
            VALUES_FIELD_NUMBER: _ClassVar[int]
            values: _containers.RepeatedCompositeFieldContainer[QuaProgram.FunctionExpression.ScalarOrVectorArgument]
            def __init__(self, values: _Optional[_Iterable[_Union[QuaProgram.FunctionExpression.ScalarOrVectorArgument, _Mapping]]] = ...) -> None: ...
        AND_FIELD_NUMBER: _ClassVar[int]
        OR_FIELD_NUMBER: _ClassVar[int]
        XOR_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        xor: QuaProgram.FunctionExpression.XorFunction
        loc: str
        def __init__(self, xor: _Optional[_Union[QuaProgram.FunctionExpression.XorFunction, _Mapping]] = ..., loc: _Optional[str] = ..., **kwargs) -> None: ...
    class BroadcastExpression(_message.Message):
        __slots__ = ("value", "loc")
        VALUE_FIELD_NUMBER: _ClassVar[int]
        LOC_FIELD_NUMBER: _ClassVar[int]
        value: QuaProgram.AnyScalarExpression
        loc: str
        def __init__(self, value: _Optional[_Union[QuaProgram.AnyScalarExpression, _Mapping]] = ..., loc: _Optional[str] = ...) -> None: ...
    class GlobalVarRefExpression(_message.Message):
        __slots__ = ("loc", "bits", "operation")
        LOC_FIELD_NUMBER: _ClassVar[int]
        BITS_FIELD_NUMBER: _ClassVar[int]
        OPERATION_FIELD_NUMBER: _ClassVar[int]
        loc: str
        bits: _containers.RepeatedScalarFieldContainer[int]
        operation: QuaProgram.GlobalVarOperation
        def __init__(self, loc: _Optional[str] = ..., bits: _Optional[_Iterable[int]] = ..., operation: _Optional[_Union[QuaProgram.GlobalVarOperation, str]] = ...) -> None: ...
    class DynamicConfig(_message.Message):
        __slots__ = ("version", "root")
        VERSION_FIELD_NUMBER: _ClassVar[int]
        ROOT_FIELD_NUMBER: _ClassVar[int]
        version: int
        root: _struct_pb2.Struct
        def __init__(self, version: _Optional[int] = ..., root: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    DYNCONFIG_FIELD_NUMBER: _ClassVar[int]
    SCRIPT_FIELD_NUMBER: _ClassVar[int]
    COMPILEROPTIONS_FIELD_NUMBER: _ClassVar[int]
    RESULTANALYSIS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_UPDATE_FIELD_NUMBER: _ClassVar[int]
    config: _inc_qua_config_pb2.QuaConfig
    dynConfig: QuaProgram.DynamicConfig
    script: QuaProgram.Script
    compilerOptions: QuaProgram.CompilerOptions
    resultAnalysis: QuaResultAnalysis
    config_update: _inc_qua_config_pb2.QuaConfig
    def __init__(self, config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ..., dynConfig: _Optional[_Union[QuaProgram.DynamicConfig, _Mapping]] = ..., script: _Optional[_Union[QuaProgram.Script, _Mapping]] = ..., compilerOptions: _Optional[_Union[QuaProgram.CompilerOptions, _Mapping]] = ..., resultAnalysis: _Optional[_Union[QuaResultAnalysis, _Mapping]] = ..., config_update: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ...) -> None: ...

class QuaResultAnalysis(_message.Message):
    __slots__ = ("version", "model")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    version: int
    model: _containers.RepeatedCompositeFieldContainer[_struct_pb2.ListValue]
    def __init__(self, version: _Optional[int] = ..., model: _Optional[_Iterable[_Union[_struct_pb2.ListValue, _Mapping]]] = ...) -> None: ...



import logging
from abc import ABCMeta
from typing import Any, List, Type, Tuple, Union, TypeVar, Optional, cast

from qm.type_hinting import Value
from qm.api.base_api import BaseApi
from qm.exceptions import QmValueError
from qm.api.models.jobs import PendingJobData
from qm.grpc.qm.pb.frontend_pb2_grpc import FrontendStub
from qm.utils.protobuf_utils import timestamp_to_datetime
from qm.api.models.server_details import ConnectionDetails
from qm.utils.general_utils import create_input_stream_name
from qm.api.models.capabilities import QopCaps, ServerCapabilities
from qm.grpc.qm.pb.job_manager_pb2_grpc import JobManagerServiceStub
from qm.api.stubs.deprecated_job_manager_stub import DeprecatedJobManagerServiceStub
from qm.grpc.qm.pb import frontend_pb2, qm_manager_pb2, job_manager_pb2, general_messages_pb2
from qm._QmJobErrors import (
    MissingJobError,
    MissingElementError,
    UnknownInputStreamError,
    ElementWithSingleInputError,
    InvalidElementCorrectionError,
    InvalidJobExecutionStatusError,
    ElementWithoutIntermediateFrequencyError,
    _handle_job_manager_error,
)

logger = logging.getLogger(__name__)


JobStubType = TypeVar("JobStubType", JobManagerServiceStub, DeprecatedJobManagerServiceStub)


class JobManagerBaseApi(BaseApi[JobStubType], metaclass=ABCMeta):
    def __init__(self, connection_details: ConnectionDetails):
        super().__init__(connection_details)
        self._frontend_stub = FrontendStub(self._channel)  # type: ignore[no-untyped-call]

    def set_element_correction(
        self, job_id: str, element_name: str, correction: general_messages_pb2.Matrix
    ) -> Tuple[float, float, float, float]:
        request = job_manager_pb2.SetElementCorrectionRequest(jobId=job_id, qeName=element_name, correction=correction)

        response: job_manager_pb2.SetElementCorrectionResponse = self._run(
            self._stub.SetElementCorrection, request, timeout=self._timeout
        )

        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            InvalidElementCorrectionError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def get_element_correction(self, job_id: str, element_name: str) -> Tuple[float, float, float, float]:
        request = job_manager_pb2.GetElementCorrectionRequest(jobId=job_id, qeName=element_name)

        response: job_manager_pb2.GetElementCorrectionResponse = self._run(
            self._stub.GetElementCorrection, request, timeout=self._timeout
        )
        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def insert_input_stream(self, job_id: str, stream_name: str, data: List[Value]) -> None:
        stream_name = create_input_stream_name(stream_name)
        request = job_manager_pb2.InsertInputStreamRequest(jobId=job_id, streamName=stream_name)

        if all(type(element) == bool for element in data):
            request.boolStreamData.CopyFrom(job_manager_pb2.BoolStreamData(data=cast(List[bool], data)))
        elif all(type(element) == int for element in data):
            request.intStreamData.CopyFrom(job_manager_pb2.IntStreamData(data=cast(List[int], data)))
        elif all(type(element) == float for element in data):
            request.fixedStreamData.CopyFrom(job_manager_pb2.FixedStreamData(data=cast(List[float], data)))
        else:
            raise QmValueError(
                f"Invalid type in data, type is '{set(type(el) for el in data)}', "
                f"accepted types are bool | int | float"
            )

        response: job_manager_pb2.InsertInputStreamResponse = self._run(
            self._stub.InsertInputStream, request, timeout=self._timeout
        )

        valid_errors = (
            MissingJobError,
            InvalidJobExecutionStatusError,
            UnknownInputStreamError,
        )
        _handle_job_manager_error(request, response, valid_errors)

    def halt(self, job_id: str) -> bool:
        request = frontend_pb2.HaltRequest(jobId=job_id)
        response: frontend_pb2.HaltResponse = self._run(self._frontend_stub.Halt, request, timeout=self._timeout)
        return response.ok

    def resume(self, job_id: str) -> bool:
        request = frontend_pb2.ResumeRequest(jobId=job_id)
        self._run(self._frontend_stub.Resume, request, timeout=self._timeout)
        return True

    def is_paused(self, job_id: str) -> bool:
        request = frontend_pb2.PausedStatusRequest(jobId=job_id)
        response: frontend_pb2.PausedStatusResponse = self._run(
            self._frontend_stub.PausedStatus, request, timeout=self._timeout
        )
        return response.isPaused

    def is_job_running(self, job_id: str) -> bool:
        request = frontend_pb2.IsJobRunningRequest(jobId=job_id)

        response: frontend_pb2.IsJobRunningResponse = self._run(
            self._frontend_stub.IsJobRunning, request, timeout=self._timeout
        )
        return response.isRunning

    def is_data_acquiring(self, job_id: str) -> frontend_pb2.IsJobAcquiringDataResponse.AcquiringStatus:
        request = frontend_pb2.IsJobAcquiringDataRequest(jobId=job_id)
        response: frontend_pb2.IsJobAcquiringDataResponse = self._run(
            self._frontend_stub.IsJobAcquiringData, request, timeout=self._timeout
        )
        return response.acquiringStatus

    def get_job_execution_status(self, job_id: str, quantum_machine_id: str) -> frontend_pb2.JobExecutionStatus:
        request = frontend_pb2.GetJobExecutionStatusRequest(jobId=job_id, quantumMachineId=quantum_machine_id)
        response: frontend_pb2.GetJobExecutionStatusResponse = self._run(
            self._frontend_stub.GetJobExecutionStatus, request, timeout=self._timeout
        )
        return response.status

    @staticmethod
    def _create_job_query_params(
        quantum_machine_id: str,
        job_id: Optional[str],
        position: Optional[int],
        user_id: Optional[str],
    ) -> frontend_pb2.JobQueryParams:
        request = frontend_pb2.JobQueryParams(quantumMachineId=quantum_machine_id)

        if position is not None:
            request.position.value = position

        if job_id is not None:
            request.jobId.CopyFrom(frontend_pb2.QueryValueMatcher(value=job_id))

        if user_id is not None:
            request.userId.CopyFrom(frontend_pb2.QueryValueMatcher(value=user_id))

        return request

    def remove_job(
        self,
        quantum_machine_id: str,
        job_id: Optional[str] = None,
        position: Optional[int] = None,
        user_id: Optional[str] = None,
    ) -> int:
        request = JobManagerBaseApi._create_job_query_params(quantum_machine_id, job_id, position, user_id)

        response: frontend_pb2.RemovePendingJobsResponse = self._run(
            self._frontend_stub.RemovePendingJobs, request, timeout=self._timeout
        )
        return response.numbersOfJobsRemoved

    def get_pending_jobs(
        self,
        quantum_machine_id: str,
        job_id: Optional[str],
        position: Optional[int],
        user_id: Optional[str],
    ) -> List[PendingJobData]:
        request = JobManagerBaseApi._create_job_query_params(quantum_machine_id, job_id, position, user_id)
        response = self._run(self._frontend_stub.GetPendingJobs, request, timeout=self._timeout)
        return [
            PendingJobData(
                job_id=job_id,
                position_in_queue=status.positionInQueue,
                time_added=timestamp_to_datetime(status.timeAdded),
                added_by=status.addedBy,
            )
            for job_id, status in response.pendingJobs.items()
        ]

    def get_running_job(self, machine_id: str) -> Optional[str]:
        request = qm_manager_pb2.GetRunningJobRequest(machineID=machine_id)
        response: qm_manager_pb2.GetRunningJobResponse = self._run(
            self._frontend_stub.GetRunningJob, request, timeout=self._timeout
        )

        if response.jobId:
            return response.jobId
        return None


class JobManagerApi(JobManagerBaseApi[JobManagerServiceStub]):
    @property
    def _stub_class(self) -> Type[JobManagerServiceStub]:
        return JobManagerServiceStub


class DeprecatedJobManagerApi(JobManagerBaseApi[DeprecatedJobManagerServiceStub]):
    @property
    def _stub_class(self) -> Type[DeprecatedJobManagerServiceStub]:
        return DeprecatedJobManagerServiceStub


def create_job_manager_from_api(
    api: BaseApi[Any], capabilities: ServerCapabilities
) -> Union[JobManagerApi, DeprecatedJobManagerApi]:
    if capabilities.supports(QopCaps.new_grpc_structure):
        return JobManagerApi(api.connection_details)
    return DeprecatedJobManagerApi(api.connection_details)



from qm.grpc.qm.grpc.v2 import common_types_pb2 as _common_types_pb2
from qm.grpc.qm.grpc.v2 import qmm_api_pb2 as _qmm_api_pb2
from qm.grpc.qm.pb import compiler_pb2 as _compiler_pb2
from qm.grpc.qm.pb import frontend_pb2 as _frontend_pb2
from qm.grpc.qm.pb import inc_qua_pb2 as _inc_qua_pb2
from qm.grpc.qm.pb import inc_qua_config_pb2 as _inc_qua_config_pb2
from qm.grpc.qm.pb import qm_manager_pb2 as _qm_manager_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class QmServiceGetConfigRequest(_message.Message):
    __slots__ = ("quantum_machine_id",)
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    def __init__(self, quantum_machine_id: _Optional[str] = ...) -> None: ...

class UpdateConfigRequest(_message.Message):
    __slots__ = ("quantum_machine_id", "config")
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    config: _inc_qua_config_pb2.QuaConfig
    def __init__(self, quantum_machine_id: _Optional[str] = ..., config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ...) -> None: ...

class UpdateConfigResponse(_message.Message):
    __slots__ = ("success", "error")
    class UpdateConfigResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class UpdateConfigResponseError(_message.Message):
        __slots__ = ("details", "config_validation_errors")
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        CONFIG_VALIDATION_ERRORS_FIELD_NUMBER: _ClassVar[int]
        details: str
        config_validation_errors: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.ConfigValidationMessage]
        def __init__(self, details: _Optional[str] = ..., config_validation_errors: _Optional[_Iterable[_Union[_qm_manager_pb2.ConfigValidationMessage, _Mapping]]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: UpdateConfigResponse.UpdateConfigResponseSuccess
    error: UpdateConfigResponse.UpdateConfigResponseError
    def __init__(self, success: _Optional[_Union[UpdateConfigResponse.UpdateConfigResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[UpdateConfigResponse.UpdateConfigResponseError, _Mapping]] = ...) -> None: ...

class QmServiceCloseRequest(_message.Message):
    __slots__ = ("quantum_machine_id",)
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    def __init__(self, quantum_machine_id: _Optional[str] = ...) -> None: ...

class QmServiceCloseResponse(_message.Message):
    __slots__ = ("success", "error")
    class QmServiceCloseResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class QmServiceCloseResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: QmServiceCloseResponse.QmServiceCloseResponseSuccess
    error: QmServiceCloseResponse.QmServiceCloseResponseError
    def __init__(self, success: _Optional[_Union[QmServiceCloseResponse.QmServiceCloseResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[QmServiceCloseResponse.QmServiceCloseResponseError, _Mapping]] = ...) -> None: ...

class QmServiceSimulateRequest(_message.Message):
    __slots__ = ("quantum_machine_id", "high_level_program", "simulate", "controller_connections", "config")
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    HIGH_LEVEL_PROGRAM_FIELD_NUMBER: _ClassVar[int]
    SIMULATE_FIELD_NUMBER: _ClassVar[int]
    CONTROLLER_CONNECTIONS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    high_level_program: _inc_qua_pb2.QuaProgram
    simulate: _frontend_pb2.ExecutionRequest.Simulate
    controller_connections: _containers.RepeatedCompositeFieldContainer[_frontend_pb2.InterOpxConnection]
    config: _inc_qua_config_pb2.QuaConfig
    def __init__(self, quantum_machine_id: _Optional[str] = ..., high_level_program: _Optional[_Union[_inc_qua_pb2.QuaProgram, _Mapping]] = ..., simulate: _Optional[_Union[_frontend_pb2.ExecutionRequest.Simulate, _Mapping]] = ..., controller_connections: _Optional[_Iterable[_Union[_frontend_pb2.InterOpxConnection, _Mapping]]] = ..., config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ...) -> None: ...

class QmServiceSimulateResponse(_message.Message):
    __slots__ = ("success", "error")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: _qmm_api_pb2.SimulationSuccess
    error: _qmm_api_pb2.SimulationError
    def __init__(self, success: _Optional[_Union[_qmm_api_pb2.SimulationSuccess, _Mapping]] = ..., error: _Optional[_Union[_qmm_api_pb2.SimulationError, _Mapping]] = ...) -> None: ...

class QmServiceJobsQueryParams(_message.Message):
    __slots__ = ("quantum_machine_id", "job_ids", "user_ids", "description", "status")
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    USER_IDS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    job_ids: _containers.RepeatedScalarFieldContainer[str]
    user_ids: _containers.RepeatedScalarFieldContainer[str]
    description: str
    status: _containers.RepeatedScalarFieldContainer[_common_types_pb2.JobExecutionStatus]
    def __init__(self, quantum_machine_id: _Optional[str] = ..., job_ids: _Optional[_Iterable[str]] = ..., user_ids: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., status: _Optional[_Iterable[_Union[_common_types_pb2.JobExecutionStatus, str]]] = ...) -> None: ...

class QmServiceGetJobsRequest(_message.Message):
    __slots__ = ("query",)
    QUERY_FIELD_NUMBER: _ClassVar[int]
    query: _qmm_api_pb2.JobsQueryParams
    def __init__(self, query: _Optional[_Union[_qmm_api_pb2.JobsQueryParams, _Mapping]] = ...) -> None: ...

class RemoveJobsRequest(_message.Message):
    __slots__ = ("query",)
    QUERY_FIELD_NUMBER: _ClassVar[int]
    query: _qmm_api_pb2.JobsQueryParams
    def __init__(self, query: _Optional[_Union[_qmm_api_pb2.JobsQueryParams, _Mapping]] = ...) -> None: ...

class RemoveJobsResponse(_message.Message):
    __slots__ = ("success", "error")
    class RemoveJobsResponseSuccess(_message.Message):
        __slots__ = ("removed_job_ids",)
        REMOVED_JOB_IDS_FIELD_NUMBER: _ClassVar[int]
        removed_job_ids: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, removed_job_ids: _Optional[_Iterable[str]] = ...) -> None: ...
    class RemoveJobsResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: RemoveJobsResponse.RemoveJobsResponseSuccess
    error: RemoveJobsResponse.RemoveJobsResponseError
    def __init__(self, success: _Optional[_Union[RemoveJobsResponse.RemoveJobsResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[RemoveJobsResponse.RemoveJobsResponseError, _Mapping]] = ...) -> None: ...

class QmServiceGetConfigResponse(_message.Message):
    __slots__ = ("success", "error")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetConfigSuccess
    error: GetConfigError
    def __init__(self, success: _Optional[_Union[GetConfigSuccess, _Mapping]] = ..., error: _Optional[_Union[GetConfigError, _Mapping]] = ...) -> None: ...

class GetConfigSuccess(_message.Message):
    __slots__ = ("config",)
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    config: _inc_qua_config_pb2.QuaConfig
    def __init__(self, config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ...) -> None: ...

class GetConfigError(_message.Message):
    __slots__ = ("details",)
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    details: str
    def __init__(self, details: _Optional[str] = ...) -> None: ...

class QmServiceCompileRequest(_message.Message):
    __slots__ = ("quantum_machine_id", "high_level_program", "config")
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    HIGH_LEVEL_PROGRAM_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    high_level_program: _inc_qua_pb2.QuaProgram
    config: _inc_qua_config_pb2.QuaConfig
    def __init__(self, quantum_machine_id: _Optional[str] = ..., high_level_program: _Optional[_Union[_inc_qua_pb2.QuaProgram, _Mapping]] = ..., config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ...) -> None: ...

class CompileResponse(_message.Message):
    __slots__ = ("success", "error")
    class CompilationSuccess(_message.Message):
        __slots__ = ("program_id", "messages")
        PROGRAM_ID_FIELD_NUMBER: _ClassVar[int]
        MESSAGES_FIELD_NUMBER: _ClassVar[int]
        program_id: str
        messages: _containers.RepeatedCompositeFieldContainer[_compiler_pb2.CompilerMessage]
        def __init__(self, program_id: _Optional[str] = ..., messages: _Optional[_Iterable[_Union[_compiler_pb2.CompilerMessage, _Mapping]]] = ...) -> None: ...
    class CompilationError(_message.Message):
        __slots__ = ("details", "messages", "config_validation_errors")
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        MESSAGES_FIELD_NUMBER: _ClassVar[int]
        CONFIG_VALIDATION_ERRORS_FIELD_NUMBER: _ClassVar[int]
        details: str
        messages: _containers.RepeatedCompositeFieldContainer[_compiler_pb2.CompilerMessage]
        config_validation_errors: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.ConfigValidationMessage]
        def __init__(self, details: _Optional[str] = ..., messages: _Optional[_Iterable[_Union[_compiler_pb2.CompilerMessage, _Mapping]]] = ..., config_validation_errors: _Optional[_Iterable[_Union[_qm_manager_pb2.ConfigValidationMessage, _Mapping]]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: CompileResponse.CompilationSuccess
    error: CompileResponse.CompilationError
    def __init__(self, success: _Optional[_Union[CompileResponse.CompilationSuccess, _Mapping]] = ..., error: _Optional[_Union[CompileResponse.CompilationError, _Mapping]] = ...) -> None: ...

class QmServiceAddToQueueRequest(_message.Message):
    __slots__ = ("quantum_machine_id", "high_level_program", "queue_position", "config")
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    HIGH_LEVEL_PROGRAM_FIELD_NUMBER: _ClassVar[int]
    QUEUE_POSITION_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    high_level_program: _inc_qua_pb2.QuaProgram
    queue_position: _frontend_pb2.QueuePosition
    config: _inc_qua_config_pb2.QuaConfig
    def __init__(self, quantum_machine_id: _Optional[str] = ..., high_level_program: _Optional[_Union[_inc_qua_pb2.QuaProgram, _Mapping]] = ..., queue_position: _Optional[_Union[_frontend_pb2.QueuePosition, _Mapping]] = ..., config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ...) -> None: ...

class AddToQueueResponse(_message.Message):
    __slots__ = ("success", "error")
    class AddToQueueResponseSuccess(_message.Message):
        __slots__ = ("job_id", "messages")
        JOB_ID_FIELD_NUMBER: _ClassVar[int]
        MESSAGES_FIELD_NUMBER: _ClassVar[int]
        job_id: str
        messages: _containers.RepeatedCompositeFieldContainer[_compiler_pb2.CompilerMessage]
        def __init__(self, job_id: _Optional[str] = ..., messages: _Optional[_Iterable[_Union[_compiler_pb2.CompilerMessage, _Mapping]]] = ...) -> None: ...
    class AddToQueueResponseError(_message.Message):
        __slots__ = ("details", "messages", "config_validation_errors")
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        MESSAGES_FIELD_NUMBER: _ClassVar[int]
        CONFIG_VALIDATION_ERRORS_FIELD_NUMBER: _ClassVar[int]
        details: str
        messages: _containers.RepeatedCompositeFieldContainer[_compiler_pb2.CompilerMessage]
        config_validation_errors: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.ConfigValidationMessage]
        def __init__(self, details: _Optional[str] = ..., messages: _Optional[_Iterable[_Union[_compiler_pb2.CompilerMessage, _Mapping]]] = ..., config_validation_errors: _Optional[_Iterable[_Union[_qm_manager_pb2.ConfigValidationMessage, _Mapping]]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: AddToQueueResponse.AddToQueueResponseSuccess
    error: AddToQueueResponse.AddToQueueResponseError
    def __init__(self, success: _Optional[_Union[AddToQueueResponse.AddToQueueResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[AddToQueueResponse.AddToQueueResponseError, _Mapping]] = ...) -> None: ...

class QmServiceAddCompiledToQueueRequest(_message.Message):
    __slots__ = ("quantum_machine_id", "program_id")
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    PROGRAM_ID_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    program_id: str
    def __init__(self, quantum_machine_id: _Optional[str] = ..., program_id: _Optional[str] = ...) -> None: ...

class AddCompiledToQueueResponse(_message.Message):
    __slots__ = ("success", "error")
    class AddCompiledToQueueResponseSuccess(_message.Message):
        __slots__ = ("job_id",)
        JOB_ID_FIELD_NUMBER: _ClassVar[int]
        job_id: str
        def __init__(self, job_id: _Optional[str] = ...) -> None: ...
    class AddCompiledToQueueResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: AddCompiledToQueueResponse.AddCompiledToQueueResponseSuccess
    error: AddCompiledToQueueResponse.AddCompiledToQueueResponseError
    def __init__(self, success: _Optional[_Union[AddCompiledToQueueResponse.AddCompiledToQueueResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[AddCompiledToQueueResponse.AddCompiledToQueueResponseError, _Mapping]] = ...) -> None: ...

class QmServiceGetJobsResponse(_message.Message):
    __slots__ = ("success", "error")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: _qmm_api_pb2.GetJobsSuccess
    error: _qmm_api_pb2.GetJobsError
    def __init__(self, success: _Optional[_Union[_qmm_api_pb2.GetJobsSuccess, _Mapping]] = ..., error: _Optional[_Union[_qmm_api_pb2.GetJobsError, _Mapping]] = ...) -> None: ...

class ResetDigitalFiltersRequest(_message.Message):
    __slots__ = ("quantum_machine_id",)
    QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_id: str
    def __init__(self, quantum_machine_id: _Optional[str] = ...) -> None: ...

class ResetDigitalFiltersResponse(_message.Message):
    __slots__ = ("success", "error")
    class ResetDigitalFiltersResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class ResetDigitalFiltersResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: ResetDigitalFiltersResponse.ResetDigitalFiltersResponseSuccess
    error: ResetDigitalFiltersResponse.ResetDigitalFiltersResponseError
    def __init__(self, success: _Optional[_Union[ResetDigitalFiltersResponse.ResetDigitalFiltersResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[ResetDigitalFiltersResponse.ResetDigitalFiltersResponseError, _Mapping]] = ...) -> None: ...



import warnings
from typing import Union, Literal, Optional, overload

from qm.type_hinting import NumberT
from qm.exceptions import QmQuaException
from qm.utils import deprecation_message
from qm.qua._dsl.streams.common import StreamEndpoints
from qm.qua._expressions import StructT, QuaExternalOutgoingStream
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.streams.external_streams import QuaStreamDirection, _declare_opnic_stream
from qm.qua._dsl.stream_processing.stream_processing import ResultStreamSource, _Configuration, _TimestampMode


def _get_stream_name(stream_id: Optional[Union[int, str]], adc_trace: Optional[bool]) -> str:
    scope = scopes_manager.program_scope
    scope.result_index += 1

    if stream_id is None:
        var = f"r{scope.result_index}"
        if adc_trace:
            var = "atr_" + var
    elif isinstance(stream_id, int):
        var = f"r{stream_id}"
    else:
        return stream_id

    return var


def declare_stream(adc_trace: Optional[bool] = None) -> "ResultStreamSource":
    """
    The old deprecated function to declare a QUA client output stream. The new function is declare_output_stream().

    For further details about the behavior of this function, please refer to the documentation of [declare_output_stream][qm.qua.declare_output_stream].

    Client output streams declared here can be written with [qm.qua.save][] or [qm.qua.send_to_stream][].
    They can also be used with ``timestamp_stream=`` and ``measure(..., adc_stream=...)``.

    Note:
        If the stream is an ADC trace, declaring it with the syntax ``declare_stream(adc_trace=True)``
        will add a buffer of length corresponding to the pulse length.

    Args:
        adc_trace: An optional boolean to indicate if the stream is an ADC trace. If not provided, it will be inferred automatically.
    """
    deprecation_message_details = "Please use declare_output_stream() instead"
    if adc_trace is not None:
        deprecation_message_details += ", and do not use the adc_trace argument, as it is no longer needed."
    else:
        adc_trace = False  # Default value

    warnings.warn(
        deprecation_message(
            "declare_stream",
            deprecated_in="1.2.4",
            removed_in="2.0.0",
            details=deprecation_message_details,
        ),
        DeprecationWarning,
        stacklevel=2,
    )

    return _declare_client_output_stream(stream_id=None, dtype=None, adc_trace=adc_trace)


def _declare_client_output_stream(
    stream_id: Optional[Union[int, str]], dtype: Optional[type], *, adc_trace: Optional[bool] = None
) -> ResultStreamSource:
    # In contrast to the deprecated declare_stream function, here stream_id can be auto generated, and adc_trace is not required.
    # We keep the adc_trace for edge cases of backwards compatibility.

    stream_name = _get_stream_name(stream_id, adc_trace)
    stream = ResultStreamSource(
        _Configuration(
            var_name=stream_name,
            timestamp_mode=_TimestampMode.Values,
            is_adc_trace=adc_trace,
            input=-1,
            auto_reshape=False,
            dtype=dtype,
        )
    )
    if stream_id is not None:
        # Register the stream only when a custom stream_id is provided. Otherwise, it is implicitly registered by internal DSL functions.
        scopes_manager.program_scope.add_stream_declaration(stream_name, stream)

    return stream


@overload
def declare_output_stream(
    target: Literal["client"] = "client",
    stream_id: Optional[Union[int, str]] = None,
    dtype: Optional[type] = None,
    *,
    adc_trace: Optional[bool] = None,
) -> ResultStreamSource:
    """
    Declare a client output stream.

    Client output streams are written with [qm.qua.save][] and [qm.qua.send_to_stream][].
    They can also be used with ``timestamp_stream=`` and ``measure(..., adc_stream=...)``.

    Args:
        target (Literal["client"]): The endpoint type. Must be ``"client"``. Defaults to ``"client"``.
        stream_id (Optional[Union[int, str]]): A unique identifier for the stream. If omitted, a name is
            generated automatically.
        dtype (Optional[type]): If provided, values sent with [qm.qua.send_to_stream][] are cast to this type
            before they are saved.
        adc_trace (Optional[bool]): If provided, marks the stream explicitly as an ADC trace. If omitted,
            ADC-trace behavior is inferred automatically when the stream is passed to ``measure(..., adc_stream=...)``.
    """
    ...


@overload
def declare_output_stream(
    target: Literal["opnic"], stream_id: int, dtype: type[StructT]
) -> QuaExternalOutgoingStream[StructT]:
    """
    Declare an outgoing OPNIC packet stream.

    OPNIC output streams carry packets defined by a type decorated with [qm.qua.qua_struct][].
    Send packets with [qm.qua.send_to_stream][].

    Args:
        target (Literal["opnic"]): The endpoint type. Must be ``"opnic"``.
        stream_id (int): A unique integer stream ID.
        dtype (type[StructT]): The packet type carried by the stream. This must be a type decorated with
            [qm.qua.qua_struct][].
    """
    ...


def declare_output_stream(
    target: StreamEndpoints = "client",
    stream_id: Optional[Union[int, str]] = None,
    dtype: Union[Optional[type[NumberT]], type[StructT]] = None,
    *,
    adc_trace: Optional[bool] = None,
) -> Union[QuaExternalOutgoingStream[StructT], ResultStreamSource]:
    """
    Declare an output stream.

    This function supports two endpoint types:

    - ``declare_output_stream("client", stream_id=None, dtype=None, adc_trace=None)`` declares a client output stream. Equivalent to the old ``declare_stream()``.
      Client output streams are written with [qm.qua.save][] and [qm.qua.send_to_stream][],
      and can also be used with ``timestamp_stream=`` and ``measure(..., adc_stream=...)``.
    - ``declare_output_stream("opnic", stream_id, PacketType)`` declares an outgoing OPNIC stream.
      OPNIC streams carry packets defined by a ``@qua_struct`` type and are written with
      [qm.qua.send_to_stream][].

    Args:
        target: The endpoint type that receives data from the stream, either ``"client"`` or ``"opnic"``.
            Defaults to ``"client"``.
        stream_id: A unique identifier for the stream. For client streams, this may be omitted, in which case a
            unique string is generated automatically. OPNIC streams require an integer stream ID.
        dtype: The data type carried by the stream. For client streams this can be used to cast values sent with
            [qm.qua.send_to_stream][]. If omitted for client streams, values are saved without additional casting.
            For OPNIC streams this must be a ``qua_struct`` type.
        adc_trace: Client streams only. If provided, marks the stream explicitly as an ADC trace. If omitted,
            ADC-trace behavior is inferred automatically when the stream is passed to
            ``measure(..., adc_stream=...)``. This argument is retained for backward compatibility.

    Returns:
        The declared output stream.

    Examples:
        ```python
        result_stream = declare_output_stream()
        timestamp_stream = declare_output_stream("client", "timestamps")

        @qua_struct
        class Packet:
            data: QuaArray[int, 1]

        outgoing_packet_stream = declare_output_stream("opnic", 2, Packet)
        ```

    """

    if target == "client":
        return _declare_client_output_stream(stream_id=stream_id, dtype=dtype, adc_trace=adc_trace)

    elif target == "opnic":
        # Validation of dtype is done inside declare_opnic_stream, therefore we ignore the mypy warning here.
        return _declare_opnic_stream(dtype, stream_id, QuaStreamDirection.OUTGOING)  # type: ignore[arg-type, return-value]

    else:
        raise QmQuaException(f"Unsupported target for output stream: {target!r}")



from enum import Enum
from collections.abc import Collection
from typing import Union, Mapping, Optional, cast

from qm.grpc.octave.v1 import api_pb2
from qm.exceptions import QmQuaException
from qm.utils.protobuf_utils import which_one_of
from qm.octave_sdk.connectivity.connectivity_util import RFInputRFSource
from qm.octave_sdk.connectivity.exceptions import (
    ConnectivityException,
    RfSourceTypeException,
    InvalidIdentityException,
    ExternalInputTypeException,
)


# Convert gRPC RfSource oneof to local enum
class RfSourceType(str, Enum):
    rf_up_conv_output = "rf_up_conv_output"
    synth_output = "synth_output"
    external_input = "external_input"
    constant_source = "constant_source"


# Convert gRPC IfSource oneof to local enum
class IfSourceType(str, Enum):
    rf_downconv_if_source = "rf_downconv_if_source"
    if_downconv_if_source = "if_downconv_if_source"
    external_if_input = "external_if_input"
    constant_source = "constant_source"


def get_rf_source_type(rf_source: api_pb2.RFSource) -> str:
    one_of, _ = which_one_of(rf_source, "source")
    return one_of


def get_if_source_type(if_source: api_pb2.IFSource) -> str:
    one_of, _ = which_one_of(if_source, "source")
    return one_of


def validate_rf_source_type(
    rf_source: api_pb2.RFSource, rf_source_types: Collection[str], message: Optional[str]
) -> None:
    one_of, value = which_one_of(rf_source, "source")
    if one_of not in rf_source_types:
        raise InvalidIdentityException(f"{message} - {one_of}")


def get_rf_source_name(rf_source: api_pb2.RFSource) -> RFInputRFSource:
    name, value = which_one_of(rf_source, "source")
    if value is None:
        raise QmQuaException("source is not valid")

    if name == "rf_up_conv_output":
        try:
            return RFInputRFSource(cast(api_pb2.UpConvRFOutput, value).index)
        except KeyError:
            raise ConnectivityException("up converter index is not valid")
    elif name == "external_input":
        external_one_of, external_value = which_one_of(value, "input")
        if external_one_of == "rf_in_index":
            assert external_value is not None
            return RFInputRFSource.RF_in
        else:
            raise ExternalInputTypeException("external input is not valid")
    elif name == "constant_source":
        return RFInputRFSource.Off
    else:
        # "synth_output" not supported
        raise RfSourceTypeException("rf source is not supported")


def get_rf_source_from_synth_panel_output(rf_source: api_pb2.RFSource) -> api_pb2.SynthRFOutput:
    one_of, value = which_one_of(rf_source, "source")

    if isinstance(value, api_pb2.SynthRFOutput):
        return value
    else:
        # "rf_up_conv_output" and "constant_source"
        # and "external_input" are not supported
        raise RfSourceTypeException("rf source is not supported")


identity_object_to_module_ref_type_mapping: Mapping[  # type: ignore[name-defined]
    Union[
        type[api_pb2.RFUpConvIdentity],
        type[api_pb2.RFDownConvIdentity],
        type[api_pb2.IFDownConvIdentity],
        type[api_pb2.SynthIdentity],
    ],
    api_pb2.OctaveModule.ValueType,
] = {
    api_pb2.RFUpConvIdentity: api_pb2.OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER,
    api_pb2.RFDownConvIdentity: api_pb2.OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER,
    api_pb2.IFDownConvIdentity: api_pb2.OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER,
    api_pb2.SynthIdentity: api_pb2.OctaveModule.OCTAVE_MODULE_SYNTHESIZER,
}



# These imports are maintained for backwards compatibility with the validation repository, despite the restriction on importing from "_dsl".
# We have to do this temporarily because our ci/cd pipeline runs the validation framework ('main' branch), so until the 'main' branch of the validation repository is updated to use the new imports, we need to keep these imports here.
# TODO: Remove these once the imports in the validation repository are corrected, which should be addressed as soon as possible.
from qm.qua._expressions import QuaVariable
from qm.qua._dsl.measure.measure_process_factories import time_tagging
from qm.qua._dsl.stream_processing.stream_processing import ResultStreamSource as _ResultSource
from qm.qua._dsl.amplitude import _PulseAmp  # This is being used by QUAM, check with Serwan before removing

_Variable = QuaVariable  # This alias is for supporting an import that appears in QUA-lang tools. TODO: Remove this alias once the import in QUA-lang tools is corrected.



from abc import ABCMeta
from typing import Union, TypeVar, Callable, Optional, Protocol, cast

from google.protobuf.empty_pb2 import Empty

from qm.exceptions import QmQuaException, QopResponseError
from qm.api.base_api import BaseApi, StubType, RequestMessageType, ResponseMessageType

SuccessType = TypeVar("SuccessType", covariant=True)


class ResponseMessageTypeV2(Protocol[SuccessType]):
    def HasField(self, name: str) -> bool:
        ...

    @property
    def success(self) -> SuccessType:
        ...


class BaseApiV2(BaseApi[StubType], metaclass=ABCMeta):
    def _run(
        self,
        grpc_method: Callable[..., ResponseMessageType],
        request: Union[RequestMessageType, Empty],
        timeout: Optional[float] = None,
    ) -> SuccessType:  # type: ignore[type-var]

        response = super()._run(grpc_method, request, timeout)
        resp = cast(ResponseMessageTypeV2[SuccessType], response)
        return self._handle_reponse(resp)

    def _handle_reponse(self, response: ResponseMessageTypeV2[SuccessType]) -> SuccessType:
        if response.HasField("success"):
            return response.success
        elif response.HasField("error"):
            raise QopResponseError(error=response.error)  # type: ignore[attr-defined]
        else:
            raise QmQuaException("got response without success or error")



from typing import Type, Tuple, Union, Optional

from google.protobuf.message import Message as PbMessage

from qm._loc import _get_loc
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.qua._scope_management.scopes import _ArbitraryScope
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.proto_any_converter import ProtoMessageT, any_to_proto, proto_to_any


def get_arbitrary_message(name: str, message: PbMessage) -> inc_qua_pb2.QuaProgram.AnyStatement:
    """
    Create an arbitrary message statement to support non-qua messages.

    Args:
        name (str): message name, used to identify the message type.
        message (PbMessage): the message needs to be encapsulated.

    Returns:
        QuaProgram.AnyStatement: The arbitrary message statement.
    """
    loc = _get_loc()
    proto_any = proto_to_any(message)
    statement = inc_qua_pb2.QuaProgram.ArbitraryStatement(
        loc=loc,
        name=name,
        data=proto_any,
    )
    return inc_qua_pb2.QuaProgram.AnyStatement(arbitrary=statement)


def arbitrary(name: str, message: PbMessage) -> None:
    """arbitrary message to support non-qua messages.

    For example, this can be used to send a protobuf message as long as the client and the server know the message type.

    Args:
        name (str): message name, used to identify the message type.
        message (PbMessage): the message needs to be encapsulated.
    """
    statement = get_arbitrary_message(name, message)
    scopes_manager.append_statement(statement)


def parse_arbitrary(b: bytes) -> inc_qua_pb2.QuaProgram.ArbitraryStatement:
    """
    Parse an arbitrary statement from bytes.

    Args:
        b (bytes): The bytes to parse.

    Returns:
        QuaProgram.ArbitraryStatement: The parsed arbitrary statement.
    """
    parsed_message = inc_qua_pb2.QuaProgram.AnyStatement()
    parsed_message.ParseFromString(b)
    if not parsed_message.HasField("arbitrary"):
        raise QmQuaException("Failed to parse arbitrary statement, expected 'arbitrary' field not found.")
    return parsed_message.arbitrary


def arbitrary_context(name: str, data: Optional[PbMessage] = None) -> _ArbitraryScope:
    """
    arbitrary context message to support non-qua contexts.

    For example, this can be used to send a protobuf message as long as the client and the server know the message type.

    Args:
        name (str): message name, used to identify the message type.
        data (Any): the message name data.

    Returns:
        Scope: An arbitrary scope that can be used to wrap statements.
    """
    if isinstance(data, PbMessage):
        data_to_context = proto_to_any(data)
    else:
        data_to_context = None
    return _ArbitraryScope(loc=_get_loc(), name=name, data=data_to_context)


def extract_from_arbitrary(
    statement: Union[inc_qua_pb2.QuaProgram.ArbitraryStatement, inc_qua_pb2.QuaProgram.ArbitraryContextStatement],
    message_cls: Type[ProtoMessageT],
) -> Tuple[str, Optional[ProtoMessageT]]:
    """
    Extracts the name and data from an arbitrary statement.

    Args:
        statement (QuaProgram.ArbitraryStatement or QuaProgram.ArbitraryContextStatement): The statement to extract from.
        message_cls (Type[ProtoMessageT]): The message class to unpack into.

    Returns:
        Tuple[str, Optional[ProtoMessageT]]: The name and the unpacked message.
    """

    message: Union[ProtoMessageT, None]
    if statement.HasField("data"):
        if issubclass(message_cls, PbMessage):
            message = any_to_proto(statement.data, message_cls)
        else:
            raise QmQuaException(f"unsupported arbitrary statement, expected PbMessage, got {message_cls}")
    else:
        message = None

    return statement.name, message
